# Copyright Open Logistics Foundation
#
# Licensed under the Open Logistics Foundation License 1.3.
# For details on the licensing terms, see the LICENSE file.
# SPDX-License-Identifier: OLFL-1.3

"""Module for parsing Label Studio formatted annotations from a single json file, which
is generated by the export method of Label Studio for one task
"""

import logging
import os
from typing import List, Optional, Tuple

import cv2

from mlcvzoo_base.api.data.annotation import BaseAnnotation
from mlcvzoo_base.api.data.annotation_class_mapper import AnnotationClassMapper
from mlcvzoo_base.api.data.annotation_parser import AnnotationParser
from mlcvzoo_base.api.data.types import FrameShape
from mlcvzoo_base.api.exceptions import ForbiddenClassError
from mlcvzoo_base.configuration.annotation_handler_config import (
    AnnotationHandlerSingleFileInputDataConfig,
)
from mlcvzoo_base.data_preparation.annotation_parser.label_studio_annotation_parser import (
    LabelStudioAnnotationData,
    get_label_studio_annotation_dict_single_file,
    parse_annotation_data,
)
from mlcvzoo_base.utils.file_utils import get_image_file_dict

logger = logging.getLogger(__name__)


class LabelStudioAnnotationParserSingle(AnnotationParser):
    """
    Parser for the Label Studio json format of a single json file that is
    generated by the export method of Label Studio for one Project

    Support data types:
    - BoundingBox
    """

    def __init__(
        self,
        mapper: AnnotationClassMapper,
        label_studio_input_data: List[AnnotationHandlerSingleFileInputDataConfig],
    ):
        AnnotationParser.__init__(self, mapper=mapper)

        self.label_studio_input_data: List[AnnotationHandlerSingleFileInputDataConfig] = (
            label_studio_input_data
        )

    def parse(self) -> List[BaseAnnotation]:
        annotations: List[BaseAnnotation] = []

        for dataset_count, input_data in enumerate(self.label_studio_input_data):
            image_file_dict = get_image_file_dict(
                input_dir=input_data.input_root_dir, image_format=input_data.image_format
            )
            label_studio_dict = get_label_studio_annotation_dict_single_file(
                label_studio_annotation_path=input_data.input_path
            )
            if label_studio_dict is None:
                continue

            if label_studio_dict is None:
                logger.warning(
                    "Label Studio annotation content could not be loaded from: '%s', "
                    "Annotations will be skipped." % input_data.input_path
                )
                continue

            if not isinstance(label_studio_dict, list):
                logger.warning(
                    "Loaded Label Studio annotation content is not of type list, file: '%s', "
                    "Annotation will be skipped." % input_data.input_path
                )
                continue

            for label_studio_task in label_studio_dict:
                label_studio_image_name = os.path.basename(label_studio_task["data"]["image"])

                if label_studio_image_name in image_file_dict:
                    image_path = image_file_dict[label_studio_image_name]
                else:
                    logger.warning(
                        "Label Studio annotation contains non-existing image: '%s', "
                        "Annotation will be skipped." % label_studio_image_name
                    )
                    continue

                try:
                    for annotation in label_studio_task["annotations"]:
                        # REMARK: image_shape will be set multiple times.
                        #         It is assumed that all annotations belong to
                        #         the same image and therefore only ONE image_shape
                        #         has to be set.

                        label_studio_annotation_data: LabelStudioAnnotationData = (
                            parse_annotation_data(
                                results=annotation["result"],
                                mapper=self.mapper,
                            )
                        )

                        if label_studio_annotation_data.image_shape is None:
                            _image_shape = FrameShape(*cv2.imread(image_path).shape)
                            label_studio_annotation_data.image_shape = (
                                _image_shape.height,
                                _image_shape.width,
                            )

                        annotations.append(
                            BaseAnnotation(
                                image_path=image_path,
                                annotation_path=input_data.input_path,
                                image_shape=label_studio_annotation_data.image_shape,
                                classifications=label_studio_annotation_data.classifications,
                                bounding_boxes=label_studio_annotation_data.bounding_boxes,
                                segmentations=label_studio_annotation_data.segmentations,
                                image_dir=os.path.dirname(image_path),
                                annotation_dir=os.path.dirname(input_data.input_path),
                                # TODO: How to set this?
                                replacement_string="",
                            )
                        )
                except (ValueError, ForbiddenClassError, KeyError) as error:
                    logger.warning("%s, annotation will be skipped" % str(error))
                    continue

        return annotations



# Copyright Open Logistics Foundation
#
# Licensed under the Open Logistics Foundation License 1.3.
# For details on the licensing terms, see the LICENSE file.
# SPDX-License-Identifier: OLFL-1.3

"""Module for different utility operations during object detection evaluation"""

import logging
import math
import os
from pathlib import Path
from typing import Any, Dict, List, Tuple

import cv2
import numpy as np
from terminaltables import AsciiTable

from mlcvzoo_base.api.data.box import Box
from mlcvzoo_base.api.data.geometric_classifiction import GeometricClassification
from mlcvzoo_base.api.data.metrics import compute_iou
from mlcvzoo_base.api.data.types import ImageType, PolygonTypeNP
from mlcvzoo_base.evaluation.geometric.data_classes import (
    CONFUSION_MATRIX_DICT_TYPE,
    METRIC_DICT_TYPE,
    METRIC_IMAGE_INFO_TYPE,
    MetricImageInfo,
)
from mlcvzoo_base.evaluation.geometric.structs import GeometricSizeTypes

logger = logging.getLogger(__name__)


def get_bbox_size_type(box: Box, image_shape: Tuple[int, int] = (1080, 1440)) -> str:
    """
    Args:
        box: Box object
        image_shape: Tuple of ints, describing the image shape

    Returns:
        a String, the category of the size of the box (small, medium or large)
    """

    bbox_height = (int(box.ymax) - int(box.ymin)) / image_shape[0] * 480
    bbox_width = (int(box.xmax) - int(box.xmin)) / image_shape[1] * 640

    bbox_sqrt_area = math.sqrt(bbox_height * bbox_width)

    small_sqrt_area_limit = 32
    large_sqrt_area_limit = 96

    if bbox_sqrt_area <= small_sqrt_area_limit:
        return str(GeometricSizeTypes.BBOX_SMALL)

    elif small_sqrt_area_limit < bbox_sqrt_area <= large_sqrt_area_limit:
        return str(GeometricSizeTypes.BBOX_MEDIUM)

    else:
        # Any other case is bbox_sqrt_area > large_sqrt_area_limit:
        return str(GeometricSizeTypes.BBOX_LARGE)


def compute_max_prediction(
    prediction: GeometricClassification,
    gt_predictions: List[GeometricClassification],
) -> Tuple[float, GeometricClassification]:
    """
    Determine the ground truth bounding box that has the highest overlap with
    the given (predicted) bounding box

    Args:
        prediction: BoundingBox object
        gt_predictions: List of BoundingBox objects

    Returns:
        A Tuple containing the ground truth bounding box wit the highest overlap and the
        according maximum overlap score (IOU metric)
    """

    overlaps = []
    for index, gt_object in enumerate(gt_predictions):
        overlaps.append(compute_iou(gt_object, prediction))

    np_overlaps = np.asarray([overlaps])

    # get the index of the gt box with the highest overlap (alongside the correct axis)
    max_gt_bbox_index = int(np.argmax(np_overlaps, axis=1))

    # 0 as index of the first element of 2D-array of shape (1, K)
    max_overlap = float(np_overlaps[0, max_gt_bbox_index])

    assigned_gt_object = gt_predictions[max_gt_bbox_index]

    return max_overlap, assigned_gt_object


def generate_metric_table(
    metrics_dict: METRIC_DICT_TYPE, iou_threshold: float, reduced: bool = False
) -> AsciiTable:
    """
    Generate a 'AsciiTable' object filled with the metrics of a object detection evaluation.
    The columns display the attributes of dataclass
    the mlcvzoo.evaluation.geometric.data_classes.GeometricMetrics.

    Args:
        metrics_dict: The dictionary containing the metrics that should be formed into a table
        iou_threshold: The iou-threshold for which the table should be generated
        reduced: Whether to use all available metrics, or only the basic ones

    Returns:
        The generated 'AsciiTable'
    """

    table_data = [
        [
            f"{'class':15s}",
            f"{'TP':6s}",
            f"{'FP':6s}",
            f"{'FN':6s}",
            f"{'CP':6s}",
            f"{'Recall':10s}",
            f"{'Precision':10s}",
            f"{'F1':10s}",
            f"{'AP':8s}",
            f"{'AVG_TP_IOU':8s}",
        ]
    ]
    row_data: List[Any]
    for bbox_size_type in GeometricSizeTypes.get_values_as_list(class_type=GeometricSizeTypes):
        if bbox_size_type != GeometricSizeTypes.BBOX_ALL:
            row_data = [f"SIZE: {bbox_size_type}", "", "", "", "", "", ""]
            table_data.append(row_data)

        for class_name, od_metric in metrics_dict[iou_threshold][bbox_size_type].items():
            if od_metric.COUNT > 0:
                row_data = [
                    class_name,
                    f"{od_metric.TP}",
                    f"{od_metric.FP}",
                    f"{od_metric.FN}",
                    f"{od_metric.COUNT}",
                    f"{od_metric.RC:.4f}",
                    f"{od_metric.PR:.4f}",
                    f"{od_metric.F1:.4f}",
                    f"{od_metric.AP:.4f}",
                    f"{od_metric.AVG_TP_IOU:.4f}",
                ]
            else:
                row_data = [
                    class_name,
                    0,
                    0,
                    0,
                    0,
                    f"{0.0:.4f}",
                    f"{0.0:.4f}",
                    f"{0.0:.4f}",
                    f"{0.0:.4f}",
                    f"{0.0:.4f}",
                ]

            table_data.append(row_data)

        if bbox_size_type != "l":
            row_data = ["", "", "", "", "", "", ""]
            table_data.append(row_data)

    table = AsciiTable(table_data)
    table.inner_footing_row_border = False

    return table


def generate_fn_fp_confusion_matrix_table_from_dict(
    confusion_matrix: CONFUSION_MATRIX_DICT_TYPE,
) -> AsciiTable:
    """
    Generate a 'AsciiTable' object that represents the confusion matrix
    of an objection detection evaluation output. The confusion matrix indicates
    which false negative bounding boxes of a certain class could be matches to
    false positive bounding boxes of another class.

    Args:
        confusion_matrix: The confusion Matrix as 2D Dict

    Returns:
        The generated 'AsciiTable'
    """

    table_data = [[""]]

    for class_identifier_str in confusion_matrix.keys():
        table_data[0].append(class_identifier_str)

    row_data: List[Any]
    for row_class_identifier_str in confusion_matrix.keys():
        row_data = [row_class_identifier_str]
        for column_class_identifier_str in confusion_matrix[row_class_identifier_str].keys():
            row_data.append(
                f"{confusion_matrix[row_class_identifier_str][column_class_identifier_str]}"
            )
        table_data.append(row_data)

    table = AsciiTable(table_data)
    table.inner_footing_row_border = False

    return table


def create_fp_fn_images(
    metrics_image_info_dict: METRIC_IMAGE_INFO_TYPE,
) -> Dict[str, Dict[str, ImageType]]:
    """

    Args:
        metrics_image_info_dict:

    Returns:
        Dict - 1st Key: Image name of the fp-fn image
               2nd Key: Class name of the fp-fn image
               Value: Image as np array
    """
    img_directory_id_dict: Dict[str, int] = dict()

    fp_fn_image_dict: Dict[str, Dict[str, ImageType]] = {}

    for class_identifier_str, image_dict in metrics_image_info_dict.items():
        for image_path, metric_image_info in image_dict.items():
            image_id, img_directory_id_dict = generate_img_id_map(
                image_path=image_path, img_directory_id_dict=img_directory_id_dict
            )

            if metric_image_info.gt_evaluation_entry is not None:
                gt_count = len(metric_image_info.gt_evaluation_entry.evaluation_objects)
            else:
                gt_count = 0

            if metric_image_info.tp_evaluation_entry is not None:
                tp_count = len(metric_image_info.tp_evaluation_entry.evaluation_objects)
            else:
                tp_count = 0

            if metric_image_info.fp_evaluation_entry is not None:
                fp_count = len(metric_image_info.fp_evaluation_entry.evaluation_objects)
            else:
                fp_count = 0

            if metric_image_info.fn_evaluation_entry is not None:
                fn_count = len(metric_image_info.fn_evaluation_entry.evaluation_objects)
            else:
                fn_count = 0

            image_basename, image_ext = os.path.splitext(os.path.basename(image_path))

            # Only create images when something False Positives or False Negatives are present
            # for this image
            if fp_count > 0 or fn_count > 0:
                fp_fn_image_name = (
                    f"metric_image_info_{class_identifier_str}/"
                    f"{image_basename}_"
                    f"ID_{image_id}_"
                    f"GT_{gt_count}_"
                    f"TP_{tp_count}_"
                    f"FP_{fp_count}_"
                    f"FN_{fn_count}_"
                    f"{image_ext}"
                )
                logger.debug(
                    "Create fp-fn image '%s' for image '%s'" % (fp_fn_image_name, image_path)
                )

                image = __create_fp_fn_image(
                    image_path=image_path,
                    metric_image_info=metric_image_info,
                )
                if fp_fn_image_name not in fp_fn_image_dict:
                    fp_fn_image_dict[fp_fn_image_name] = {}

                fp_fn_image_dict[fp_fn_image_name][class_identifier_str] = image

    return fp_fn_image_dict


def __draw_polygon(
    image: ImageType, polygon: PolygonTypeNP, color: Tuple[int, int, int], thickness: int
) -> ImageType:
    pts = np.array(polygon, np.int32)  # type: ignore[var-annotated]
    pts = pts.reshape((-1, 1, 2))

    image = cv2.polylines(img=image, pts=[pts], isClosed=True, color=color, thickness=thickness)

    return image


def __create_fp_fn_image(image_path: str, metric_image_info: MetricImageInfo) -> ImageType:
    image = cv2.imread(image_path)

    tp_color = (0, 255, 0)  # => red color for false positive boxes
    fp_color = (0, 0, 255)  # => red color for false positive boxes
    fn_color = (255, 0, 0)  # => black color for false negative boxes
    gt_color = (255, 255, 255)  # => white color for ground truth data

    if metric_image_info.gt_evaluation_entry is not None:
        for geometric_object in metric_image_info.gt_evaluation_entry.evaluation_objects:
            image = __draw_polygon(
                image=image, polygon=geometric_object.polygon(), color=gt_color, thickness=6
            )
    if metric_image_info.tp_evaluation_entry is not None:
        for geometric_object in metric_image_info.tp_evaluation_entry.evaluation_objects:
            image = __draw_polygon(
                image=image, polygon=geometric_object.polygon(), color=tp_color, thickness=6
            )

    if metric_image_info.fn_evaluation_entry is not None:
        for geometric_object in metric_image_info.fn_evaluation_entry.evaluation_objects:
            image = __draw_polygon(
                image=image, polygon=geometric_object.polygon(), color=fn_color, thickness=3
            )

    if metric_image_info.fp_evaluation_entry is not None:
        for geometric_object in metric_image_info.fp_evaluation_entry.evaluation_objects:
            image = __draw_polygon(
                image=image, polygon=geometric_object.polygon(), color=fp_color, thickness=3
            )

    return image  # type: ignore[no-any-return]


def generate_img_id_map(
    image_path: str, img_directory_id_dict: Dict[str, int]
) -> Tuple[int, Dict[str, int]]:
    """
    Generates an index (id) for the given image_path.

    Args:
        image_path: String, path to an image
        img_directory_id_dict: Dictionary of image paths to index in directory list

    Returns: a Tuple of image index and (updated) img_directory_id_dict
    """
    # TODO: What happens when two images get the same index?
    #  It could be possible that due to an extension of the dict two images could get the same index,
    #  because the indices after the added and sorted image are increased by 1.
    #  Therefore all other entries of the subsequent images (in sorted order) are wrong.

    image_dir = os.path.dirname(image_path)
    file_extension = os.path.splitext(image_path)[1]

    if image_path not in img_directory_id_dict:
        path_generator = Path(image_dir).glob(f"**/*{file_extension}")

        file_list = [str(p) for p in path_generator]

        file_list.sort()

        for index, file_path in enumerate(file_list):
            img_directory_id_dict[file_path] = index

    return img_directory_id_dict[image_path], img_directory_id_dict



# Copyright Open Logistics Foundation
#
# Licensed under the Open Logistics Foundation License 1.3.
# For details on the licensing terms, see the LICENSE file.
# SPDX-License-Identifier: OLFL-1.3

"""Module for different utility operations regarding time and json operations"""
import inspect
import json
from datetime import datetime, timedelta
from typing import Any


def get_current_timestamp(format_string: str = "%d-%b-%Y (%H:%M:%S.%f)") -> str:
    """
    Returns the current time in the given format
    Args:
        format_string: String, defining the format of the timestamp

    Returns: String, the formatted timestamp

    """

    timestamp = datetime.now()

    timestamp_string = timestamp.strftime(format_string)

    return timestamp_string


def timestamp_to_datetime(datetime_object: datetime) -> timedelta:
    """
    Transforms a timestamp to a timedelta based on the given object's attributes

    Args:
        datetime_object: datetime, a timestamp

    Returns: a timedelta object

    """

    return timedelta(
        hours=datetime_object.hour,
        minutes=datetime_object.minute,
        seconds=datetime_object.second,
        microseconds=datetime_object.microsecond,
    )


class CustomJSONEncoder(json.JSONEncoder):
    """Class for encoding json objects"""

    def default(self, obj: Any) -> Any:
        """
        Sets the class's default attribute to the json version of the given object

        Args:
            obj: Any, an object that should be transformed to json format

        Returns: Any, the given object without changes

        """

        if hasattr(obj, "to_json"):
            return self.default(obj.to_json())
        elif hasattr(obj, "__dict__"):
            d = dict(
                (key, value)
                for key, value in inspect.getmembers(obj)
                if not key.startswith("__")
                and not inspect.isabstract(value)
                and not inspect.isbuiltin(value)
                and not inspect.isfunction(value)
                and not inspect.isgenerator(value)
                and not inspect.isgeneratorfunction(value)
                and not inspect.ismethod(value)
                and not inspect.ismethoddescriptor(value)
                and not inspect.isroutine(value)
            )
            return self.default(d)
        return obj



# Copyright Open Logistics Foundation
#
# Licensed under the Open Logistics Foundation License 1.3.
# For details on the licensing terms, see the LICENSE file.
# SPDX-License-Identifier: OLFL-1.3

"""
Module for different utility operations during data preparation
"""
import logging
import os
import xml.dom.minidom as mmd
import xml.etree.ElementTree as ET_xml
from typing import List, Optional

from mlcvzoo_base.api.data.annotation import BaseAnnotation
from mlcvzoo_base.utils.file_utils import ensure_dir
from mlcvzoo_base.utils.implicit_path_replacements import ImplicitReplacement
from mlcvzoo_base.utils.xml_utils import create_bbox_xml_entry, create_xml_root

logger = logging.getLogger(__name__)


def replace_index_by_dir(annotation: BaseAnnotation) -> BaseAnnotation:
    """
    Replaces replacement_string parts in image_path and annotation_path attributes of
    an annotation by respective directory information (image_dir and annotation_dir)

    Args:
        annotation: BaseAnnotation object

    Returns: BaseAnnotation, the input object with different path attributes

    """

    if annotation.replacement_string is not None:
        # TODO: use directory replacement
        annotation.image_path = ImplicitReplacement.replace_string_in_path(
            file_path=annotation.image_path,
            value=annotation.replacement_string,
            replacement_value=annotation.image_dir,
            how=ImplicitReplacement.FIRST,
        )

        # TODO: use directory replacement
        annotation.annotation_path = ImplicitReplacement.replace_string_in_path(
            file_path=annotation.annotation_path,
            value=annotation.replacement_string,
            replacement_value=annotation.annotation_dir,
            how=ImplicitReplacement.FIRST,
        )

        return ensure_abspath(annotation=annotation)


def replace_dir_by_index(annotation: BaseAnnotation) -> BaseAnnotation:
    """
    Replaces directory parts (image_dir and annotation_dir) in image_path and
    annotation_path attributes of an annotation by replacement_string information

    Args:
        annotation: BaseAnnotation object

    Returns: BaseAnnotation, the input object with different path attributes

    """

    if annotation.replacement_string is not None:
        annotation.image_path = ImplicitReplacement.replace_string_in_path(
            file_path=annotation.image_path,
            value=annotation.image_dir,
            replacement_value=annotation.replacement_string,
            how=ImplicitReplacement.FIRST,
        )

        annotation.annotation_path = ImplicitReplacement.replace_string_in_path(
            file_path=annotation.annotation_path,
            value=annotation.annotation_dir,
            replacement_value=annotation.replacement_string,
            how=ImplicitReplacement.FIRST,
        )

    return annotation


def ensure_abspath(annotation: BaseAnnotation) -> BaseAnnotation:
    """
    Sets the given annotation's path and directory attributes to absolute paths

    Args:
        annotation: BaseAnnotation object

    Returns: BaseAnnotation, the input object with different path attributes

    """

    annotation.image_path = os.path.abspath(annotation.image_path)
    annotation.image_dir = os.path.abspath(annotation.image_dir)
    annotation.annotation_path = os.path.abspath(annotation.annotation_path)
    annotation.annotation_dir = os.path.abspath(annotation.annotation_dir)

    return annotation


def annotation_to_xml(
    annotation: BaseAnnotation,
    allowed_classes: Optional[List[str]] = None,
) -> None:
    """
    Transforms the given annotation object to XML format and saves it to
    a file at annotation_path location

    Args:
        annotation: BaseAnnotation object
        allowed_classes: Optional list of strings (allowed classes)

    Returns: None

    """

    root = create_xml_root(
        file_path=annotation.image_path,
        width=annotation.image_shape[1],
        height=annotation.image_shape[0],
    )

    assert annotation.image_shape[0] > 0
    assert annotation.image_shape[1] > 0

    if len(annotation.get_bounding_boxes(include_segmentations=True)) == 0:
        logger.warning("annotation has no bounding_boxes!")

    for bounding_box in annotation.get_bounding_boxes(include_segmentations=True):
        if allowed_classes is not None and bounding_box.class_name not in allowed_classes:
            continue

        root = create_bbox_xml_entry(
            root,
            bounding_box.class_name,
            bounding_box.ortho_box().xmin,
            bounding_box.ortho_box().ymin,
            bounding_box.ortho_box().xmax,
            bounding_box.ortho_box().ymax,
        )

    logger.info("Write xml file: %s", annotation.annotation_path)

    ensure_dir(file_path=annotation.annotation_path, verbose=True)

    tree_string = "".join(ET_xml.tostring(element=root).decode("utf-8").split())
    pretty_tree_string = mmd.parseString(tree_string).toprettyxml(
        indent="  ",
    )

    with open(file=annotation.annotation_path, mode="w", encoding="utf8") as annotation_file:
        annotation_file.write(pretty_tree_string)


def save_bbox_snippets(annotation: BaseAnnotation, output_dir: str) -> None:
    raise NotImplementedError("Currently not implemented!")
    # TODO:
    # image_path = self.image_path
    #
    # snippet_base_dir = os.path.join(
    #     output_dir,
    #     "snippets"
    # )
    #
    # for index, d in enumerate(self.bounding_boxes):
    #     snippet_name = os.path.basename(image_path)[:-4] + "_" + \
    #                    d.class_name + "_" + str(index) + os.path.basename(image_path)[-4:]
    #
    #     snippet_dir = os.path.join(snippet_base_dir, d.class_name)
    #
    #     if not os.path.exists(snippet_dir):
    #         os.makedirs(snippet_dir)
    #
    #         print("Create directory: {}".format(snippet_dir))
    #
    #     snippet_path = os.path.join(snippet_dir, snippet_name)
    #
    #     if not os.path.isfile(snippet_path):
    #         img = cv2.imread(image_path)
    #
    #         crop_img = img[d.box.ymin:d.box.ymax, d.box.xmin:d.box.xmax]
    #
    #         # NOTE: comment in to show image
    #         # cv2.imshow(snippet_name, crop_img)
    #         # cv2.waitKey(0)
    #
    #         print("Write snippet: {}".format(snippet_path))
    #         cv2.imwrite(snippet_path, crop_img)


def filter_annotation(
    annotation: BaseAnnotation,
    class_ids: Optional[List[int]] = None,
    class_names: Optional[List[str]] = None,
    classification_score: float = 0.0,
    bounding_box_score: float = 0.0,
    segmentation_score: float = 0.0,
) -> BaseAnnotation:
    """
    Filter a given annotation object. Afterwards the object only contains
    the classifications, bounding-boxes and segmentations that fulfill:

    - the class id matches one of the given ids given by 'class_ids' (only when the list is
      not None nor empty)
    - the class name matches one of the given names given by 'class_names' (only when the list is
      not None nor empty)
    - the score is higher than the one associated with this datastructure

    Args:
        annotation: the annotation object to filter
        class_ids: the class ids that are valid
        class_names: the class names that are valid
        classification_score: the score a any Classification object has to fulfill
        bounding_box_score: the score a any BoundingBox object has to fulfill
        segmentation_score: the score a any Segmentation object has to fulfill

    Returns:
        The filter annotation object
    """

    filtered_annotation: BaseAnnotation = BaseAnnotation(
        image_path=annotation.image_path,
        annotation_path=annotation.annotation_path,
        image_shape=annotation.image_shape,
        image_dir=annotation.image_dir,
        replacement_string=annotation.replacement_string,
        annotation_dir=annotation.annotation_dir,
        classifications=[],
        bounding_boxes=[],
        segmentations=[],
    )

    for classification in annotation.classifications:
        if (
            class_ids is not None
            and len(class_ids) > 0
            and classification.class_id not in class_ids
        ) or (
            class_names is not None
            and len(class_names) > 0
            and classification.class_name not in class_names
        ):
            continue

        if classification.score >= classification_score:
            filtered_annotation.classifications.append(classification)

    for bounding_box in annotation.bounding_boxes:
        if (
            class_ids is not None and len(class_ids) > 0 and bounding_box.class_id not in class_ids
        ) or (
            class_names is not None
            and len(class_names) > 0
            and bounding_box.class_name not in class_names
        ):
            continue

        if bounding_box.score >= bounding_box_score:
            filtered_annotation.bounding_boxes.append(bounding_box)

    for segmentation in annotation.segmentations:
        if (
            class_ids is not None and len(class_ids) > 0 and segmentation.class_id not in class_ids
        ) or (
            class_names is not None
            and len(class_names) > 0
            and segmentation.class_name not in class_names
        ):
            continue

        if segmentation.score >= segmentation_score:
            filtered_annotation.segmentations.append(segmentation)

    return filtered_annotation


def filter_annotations(
    annotations: List[BaseAnnotation],
    class_ids: Optional[List[int]] = None,
    class_names: Optional[List[str]] = None,
    classification_score: float = 0.0,
    bounding_box_score: float = 0.0,
    segmentation_score: float = 0.0,
) -> List[BaseAnnotation]:
    """
    Filter a given list of annotations. Afterwards each annotation only contains
    the classifications, bounding-boxes and segmentations that fulfill:

    - the class id matches one of the given ids given by 'class_ids' (only when the list is
      not None nor empty)
    - the class name matches one of the given names given by 'class_names' (only when the list is
      not None nor empty)
    - the score is higher than the one associated with this datastructure

    Args:
        annotations: the list of annotations to filter
        class_ids: the class ids that are valid
        class_names: the class names that are valid
        classification_score: the score a any Classification object has to fulfill
        bounding_box_score: the score a any BoundingBox object has to fulfill
        segmentation_score: the score a any Segmentation object has to fulfill

    Returns:
        The filtered list of annotations
    """

    filtered_annotations: List[BaseAnnotation] = []

    for annotation in annotations:
        filtered_annotations.append(
            filter_annotation(
                annotation=annotation,
                class_ids=class_ids,
                class_names=class_names,
                classification_score=classification_score,
                bounding_box_score=bounding_box_score,
                segmentation_score=segmentation_score,
            )
        )

    return filtered_annotations



# Copyright Open Logistics Foundation
#
# Licensed under the Open Logistics Foundation License 1.3.
# For details on the licensing terms, see the LICENSE file.
# SPDX-License-Identifier: OLFL-1.3

"""Module for parsing PascalVOC formatted annotations"""
import logging
import os
from typing import List, Tuple
from xml.etree.ElementTree import ParseError

from mlcvzoo_base.api.data.annotation import BaseAnnotation
from mlcvzoo_base.api.data.annotation_class_mapper import AnnotationClassMapper
from mlcvzoo_base.api.data.annotation_parser import AnnotationParser
from mlcvzoo_base.api.exceptions import ForbiddenClassError
from mlcvzoo_base.configuration.annotation_handler_config import (
    AnnotationHandlerPASCALVOCInputDataConfig,
)
from mlcvzoo_base.configuration.structs import AnnotationFileFormats, ImageFileFormats
from mlcvzoo_base.data_preparation.annotation_builder.pascal_voc_annotation_builder import (
    PascalVOCAnnotationBuilder,
)
from mlcvzoo_base.utils.file_utils import get_file_list

logger = logging.getLogger(__name__)


class PascalVOCAnnotationParser(AnnotationParser):
    """
    Super class for defining the methods that are needed to parse a list of
    instances that are of the type BaseAnnotation.
    Each annotation format e.g. Pascal-VOC, COCO, CVAT-for-images should have
    its own child AnnotationHandler class
    """

    def __init__(
        self,
        mapper: AnnotationClassMapper,
        pascal_voc_input_data: List[AnnotationHandlerPASCALVOCInputDataConfig],
    ):
        AnnotationParser.__init__(self, mapper=mapper)

        self.pascal_voc_input_data = pascal_voc_input_data

    def parse(self) -> List[BaseAnnotation]:
        annotations: List[BaseAnnotation] = list()
        # Fill dictionary for implicit path replacement
        for dataset_count, input_data_config in enumerate(self.pascal_voc_input_data):
            replacement_string = AnnotationParser.csv_directory_replacement_string.format(
                dataset_count
            )

            image_path_list, annotation_path_list = self.__determine_data_paths(
                input_data_config=input_data_config
            )

            if input_data_config.ignore_missing_images:
                image_path_list = len(annotation_path_list) * [""]
            else:
                (
                    image_path_list,
                    annotation_path_list,
                ) = self.__ensure_annotations_match_images(
                    image_path_list=image_path_list,
                    annotation_path_list=annotation_path_list,
                )

            for image_path, annotation_path in zip(image_path_list, annotation_path_list):
                if (
                    not input_data_config.ignore_missing_images
                    and os.path.realpath(input_data_config.input_image_dir)
                    in os.path.realpath(image_path)
                    and os.path.realpath(input_data_config.input_xml_dir)
                    in os.path.realpath(annotation_path)
                ) or (
                    input_data_config.ignore_missing_images
                    and os.path.realpath(input_data_config.input_xml_dir)
                    in os.path.realpath(annotation_path)
                ):
                    try:
                        pascal_voc_builder = PascalVOCAnnotationBuilder(
                            mapper=self.mapper,
                            use_difficult=input_data_config.use_difficult,
                            use_occluded=input_data_config.use_occluded,
                            use_background=input_data_config.use_background,
                        )

                        annotation: BaseAnnotation = pascal_voc_builder.build(
                            image_path=image_path,
                            annotation_path=annotation_path,
                            image_dir=input_data_config.input_image_dir,
                            annotation_dir=input_data_config.input_xml_dir,
                            replacement_string=replacement_string,
                        )

                        annotations.append(annotation)

                    except (ForbiddenClassError, ParseError) as error:
                        logger.warning(
                            "%s: %s, annotation '%s' will be skipped"
                            % (type(error), str(error), annotation_path),
                        )
                        continue

        return annotations

    @staticmethod
    def __determine_data_paths(
        input_data_config: AnnotationHandlerPASCALVOCInputDataConfig,
    ) -> Tuple[List[str], List[str]]:
        """
        Determine all image and annotations paths from the given input-data-config

        Args:
            input_data_config: The AnnotationHandlerPASCALVOCInputDataConfig that defines
                               the input-data

        Returns:
            Tuple of image paths and annotation paths
        """

        image_paths = []
        xml_paths = []

        # CHECK SUBDIR CONFIG
        if len(input_data_config.input_sub_dirs) > 0:
            for sub_input_dir in input_data_config.input_sub_dirs:
                # READ IMAGES
                input_image_dir = os.path.join(input_data_config.input_image_dir, sub_input_dir)

                image_paths.extend(
                    get_file_list(
                        input_dir=input_image_dir,
                        search_subfolders=True,
                        file_extension=f"{input_data_config.image_format}",
                    )
                )

                # READ ANNOTATIONS
                input_xml_dir = os.path.join(input_data_config.input_xml_dir, sub_input_dir)

                xml_paths.extend(
                    get_file_list(
                        input_dir=input_xml_dir,
                        search_subfolders=True,
                        # exclude_pattern="predicted",
                        file_extension=AnnotationFileFormats.XML,
                    )
                )
        else:
            # READ IMAGES
            image_paths.extend(
                get_file_list(
                    input_dir=input_data_config.input_image_dir,
                    search_subfolders=True,
                    file_extension=input_data_config.image_format,
                )
            )

            # READ ANNOTATIONS
            xml_paths.extend(
                get_file_list(
                    input_dir=input_data_config.input_xml_dir,
                    search_subfolders=True,
                    # exclude_pattern="predicted",
                    file_extension=AnnotationFileFormats.XML,
                )
            )

        image_paths.sort()
        xml_paths.sort()

        return image_paths, xml_paths

    @staticmethod
    def __ensure_annotations_match_images(
        image_path_list: List[str], annotation_path_list: List[str]
    ) -> Tuple[List[str], List[str]]:
        """

        Args:
            image_path_list:
            annotation_path_list:

        Returns:

        """
        # remove duplicate
        image_path_list = list(dict.fromkeys(image_path_list))
        annotation_path_list = list(dict.fromkeys(annotation_path_list))

        # TODO: allow more image formats
        cleaned_img = [
            os.path.basename(image_path)
            .replace(ImageFileFormats.JPEG, "")
            .replace(ImageFileFormats.PNG, "")
            for image_path in image_path_list
        ]
        cleaned_xml = [
            os.path.basename(annotation_path)
            .replace(AnnotationFileFormats.XML, "")
            .replace("_predicted", "")
            for annotation_path in annotation_path_list
        ]

        in_both_lists = set(cleaned_xml) & set(cleaned_img)

        img_indices = [idx for (idx, x) in enumerate(cleaned_img) if x in in_both_lists]
        xml_indices = [idx for (idx, x) in enumerate(cleaned_xml) if x in in_both_lists]

        cleaned_image_path_list = [
            image_path
            for (img_index, image_path) in enumerate(image_path_list)
            if img_index in img_indices
        ]
        cleaned_annotation_path_list = [
            annotation_path
            for (annotation_index, annotation_path) in enumerate(annotation_path_list)
            if annotation_index in xml_indices
        ]

        if len(image_path_list) > len(cleaned_image_path_list):
            logger.debug(
                "Did not find an annotation-paths for image-paths: \n\t- %s\n",
                "\n\t- ".join(
                    [
                        image_path_list[index]
                        for index in range(0, len(image_path_list))
                        if index not in img_indices
                    ]
                ),
            )

        if len(annotation_path_list) > len(cleaned_annotation_path_list):
            logger.debug(
                "Did not find an image-paths for annotation-paths: \n\t- %s\n",
                "\n\t- ".join(
                    [
                        annotation_path_list[index]
                        for index in range(0, len(annotation_path_list))
                        if index not in xml_indices
                    ]
                ),
            )

        return cleaned_image_path_list, cleaned_annotation_path_list



# Copyright Open Logistics Foundation
#
# Licensed under the Open Logistics Foundation License 1.3.
# For details on the licensing terms, see the LICENSE file.
# SPDX-License-Identifier: OLFL-1.3

"""Module for configuring machine learning models"""

from abc import ABC
from typing import Optional

import related
from attr import define
from config_builder import BaseConfigClass


@define
class ModelConfiguration(BaseConfigClass, ABC):
    """
    A model configuration.
    Typically, subclasses of a configuration parser implement
    this class to provide a mechanism to feed parameters into models or training.
    """

    unique_name: str = related.StringField()


@define
class InferenceConfig(BaseConfigClass):
    """
    Overall inference config that defines parameters that
    are valid for every model that is performing an inference.
    """

    config_path: Optional[str] = related.ChildField(cls=str)
    checkpoint_path: str = related.StringField()
    score_threshold: float = related.FloatField()



# Copyright Open Logistics Foundation
#
# Licensed under the Open Logistics Foundation License 1.3.
# For details on the licensing terms, see the LICENSE file.
# SPDX-License-Identifier: OLFL-1.3

"""Module for parsing CSV formatted annotations"""
import logging
import os
from dataclasses import field
from typing import List, Optional, Tuple

from tqdm import tqdm

from mlcvzoo_base.api.data.annotation import BaseAnnotation
from mlcvzoo_base.api.data.annotation_class_mapper import AnnotationClassMapper
from mlcvzoo_base.api.data.annotation_parser import AnnotationParser
from mlcvzoo_base.api.exceptions import ForbiddenClassError
from mlcvzoo_base.configuration.annotation_handler_config import (
    AnnotationHandlerPASCALVOCInputDataConfig,
    AnnotationHandlerSingleFileInputDataConfig,
)
from mlcvzoo_base.data_preparation.annotation_builder.csv_annotation_builder import (
    CSVAnnotationBuilder,
)

logger = logging.getLogger(__name__)


class CSVAnnotationParser(AnnotationParser):
    """
    Super class for defining the methods that are needed to parse a list of
    instances that are of the type BaseAnnotation.
    Each annotation format e.g. Pascal-VOC, COCO, CVAT-for-images should have
    its own child AnnotationHandler class
    """

    def __init__(
        self,
        mapper: AnnotationClassMapper,
        csv_file_path: str,
        pascal_voc_input_data: List[AnnotationHandlerPASCALVOCInputDataConfig] = field(
            default_factory=lambda: []
        ),
        coco_input_data: List[AnnotationHandlerSingleFileInputDataConfig] = field(
            default_factory=lambda: []
        ),
        cvat_input_data: List[AnnotationHandlerSingleFileInputDataConfig] = field(
            default_factory=lambda: []
        ),
    ):
        AnnotationParser.__init__(self, mapper=mapper)

        self.csv_file_path = csv_file_path
        self.pascal_voc_input_data = pascal_voc_input_data
        self.coco_input_data = coco_input_data
        self.cvat_input_data = cvat_input_data

    def parse(self) -> List[BaseAnnotation]:
        annotations: List[BaseAnnotation] = list()

        if os.path.isfile(self.csv_file_path):
            with open(file=self.csv_file_path, mode="r", encoding="'utf-8") as annotations_file:
                csv_lines = annotations_file.readlines()

                for index in tqdm(range(len(csv_lines))):
                    csv_line = csv_lines[index].strip()

                    pascal_voc_input_data_index: Optional[int] = None
                    coco_input_data_index: Optional[int] = None
                    cvat_input_data_index: Optional[int] = None

                    pascal_voc_replacement_string: Optional[str] = None
                    coco_replacement_string: Optional[str] = None
                    cvat_replacement_string: Optional[str] = None

                    if len(self.pascal_voc_input_data) > 0:
                        (
                            pascal_voc_input_data_index,
                            pascal_voc_replacement_string,
                        ) = CSVAnnotationParser.__get_input_data_index(
                            dataset_len=len(self.pascal_voc_input_data),
                            csv_line=csv_line,
                        )

                    if len(self.coco_input_data) > 0:
                        (
                            coco_input_data_index,
                            coco_replacement_string,
                        ) = CSVAnnotationParser.__get_input_data_index(
                            dataset_len=len(self.coco_input_data),
                            csv_line=csv_line,
                        )

                    if len(self.cvat_input_data) > 0:
                        (
                            cvat_input_data_index,
                            cvat_replacement_string,
                        ) = CSVAnnotationParser.__get_input_data_index(
                            dataset_len=len(self.cvat_input_data),
                            csv_line=csv_line,
                        )

                    image_dir: Optional[str] = None
                    annotation_dir: Optional[str] = None
                    replacement_string: Optional[str] = None

                    if pascal_voc_input_data_index is not None:
                        image_dir = self.pascal_voc_input_data[
                            pascal_voc_input_data_index
                        ].input_image_dir

                        annotation_dir = self.pascal_voc_input_data[
                            pascal_voc_input_data_index
                        ].input_xml_dir

                        replacement_string = pascal_voc_replacement_string

                    elif coco_input_data_index is not None:
                        image_dir = self.coco_input_data[coco_input_data_index].input_root_dir
                        annotation_dir = self.coco_input_data[coco_input_data_index].input_path

                        replacement_string = coco_replacement_string

                    elif cvat_input_data_index is not None:
                        image_dir = self.cvat_input_data[cvat_input_data_index].input_root_dir
                        annotation_dir = self.cvat_input_data[cvat_input_data_index].input_path

                        replacement_string = cvat_replacement_string

                    if (
                        image_dir is not None
                        and annotation_dir is not None
                        and replacement_string is not None
                    ):
                        csv_builder = CSVAnnotationBuilder(csv_line=csv_line, mapper=self.mapper)

                        try:
                            annotation = csv_builder.build(
                                image_path="",
                                annotation_path="",
                                image_dir=image_dir,
                                annotation_dir=annotation_dir,
                                replacement_string=replacement_string,
                            )

                            annotations.append(annotation)

                        except (ValueError, ForbiddenClassError) as error:
                            logger.warning("%s, annotation will be skipped" % str(error))
                            continue
                    else:
                        logger.warning(
                            "Could not find valid replacement for image or annotation directory!\n"
                            "csv-line: %s \n"
                            "pascal-voc-len: %s \n"
                            "coco-len: %s \n"
                            "image_dir: %s \n"
                            "annotation_dir: %s \n"
                            % (
                                csv_line,
                                len(self.pascal_voc_input_data),
                                len(self.coco_input_data),
                                image_dir,
                                annotation_dir,
                            )
                        )
        else:
            error_string = f"ERROR: file '{self.csv_file_path}' does not exist!"

            logger.error(error_string)
            raise ValueError(error_string)

        return annotations

    @staticmethod
    def __get_input_data_index(
        dataset_len: int, csv_line: str
    ) -> Tuple[Optional[int], Optional[str]]:
        input_data_index: Optional[int] = None
        replacement_string: Optional[str] = None

        for i in range(0, dataset_len):
            replacement_string = AnnotationParser.csv_directory_replacement_string.format(i)

            logger.debug(
                "Check line: \n - replacement-string: %s \n - csv_line: %s ",
                replacement_string,
                csv_line,
            )

            if replacement_string is not None:
                image_path = csv_line.split(" ")[0]

                image_path_split = image_path.split("/")

                if image_path_split[0] == replacement_string:
                    input_data_index = i
                    break

        return input_data_index, replacement_string



# Copyright Open Logistics Foundation
#
# Licensed under the Open Logistics Foundation License 1.3.
# For details on the licensing terms, see the LICENSE file.
# SPDX-License-Identifier: OLFL-1.3

"""Module for writing MMPretrain formatted annotations."""

import logging
import os
from typing import List, Optional

from mlcvzoo_base.api.data.annotation import BaseAnnotation
from mlcvzoo_base.api.data.annotation_writer import AnnotationWriter
from mlcvzoo_base.utils import encoding_safe_imwrite

logger = logging.getLogger(__name__)


class ClassificationAnnotationWriter(AnnotationWriter):
    """
    Class for writing classification based annotations.
    """

    def __init__(
        self,
        output_file_path: str,
        data_root_dir: str,
        seperator: str,
    ):
        self.output_file_path = output_file_path
        self.data_root_dir = data_root_dir
        self.seperator = seperator

    def write(self, annotations: List[BaseAnnotation]) -> Optional[str]:

        os.makedirs(os.path.dirname(self.output_file_path), exist_ok=True)

        logger.info("Write annotations to %s", self.output_file_path)

        with open(self.output_file_path, "w") as annotation_file:
            for annotation in annotations:
                for classification in annotation.classifications:
                    annotation_file.write(
                        f"{os.path.relpath(annotation.image_path, self.data_root_dir)}"
                        f"{self.seperator}"
                        f"{classification.class_id}\n"
                    )

        return self.output_file_path


class TextCropAnnotationWriter(AnnotationWriter):
    """
    Class for writing annotations for cropped text images.
    """

    def __init__(
        self,
        output_file_path: str,
        data_root_dir: str,
        seperator: str,
        rel_image_output_dir: str,
        padding_ratio: float = 0.1,
    ):
        self.output_file_path: str = output_file_path
        self.data_root_dir: str = data_root_dir
        self.seperator: str = seperator
        self.rel_image_output_dir = rel_image_output_dir
        self.padding_ratio: float = padding_ratio

    @staticmethod
    def __create_cropped_image_filepath(
        image_path: str,
        content: str,
        class_name: str,
        index: int,
        image_output_dir: str,
    ) -> str:
        """Create image path for cropped image.

        Args:
            image_path: The annotations to crop
            content: The bounding box or segmentation to crop
            class_name: The class name for the cropped image
            index: The index of the cropped image
            image_output_dir: The directory to save cropped images to,
                if not given save cropped image to original image path

        Returns:
            The created image path
        """
        image_base_name = os.path.basename(image_path)

        image_name_split = image_base_name.split(".")
        assert len(image_name_split) == 2
        return os.path.join(
            image_output_dir,
            # Replace spaces with underscores to avoid failures
            image_name_split[0].replace(" ", "_"),
            f"{class_name}_{index}_{content}.jpg",
        )

    def write(self, annotations: List[BaseAnnotation]) -> Optional[str]:

        os.makedirs(os.path.dirname(self.output_file_path), exist_ok=True)

        logger.info("Write annotations to %s", self.output_file_path)

        with open(self.output_file_path, "w") as annotation_file:
            for annotation in annotations:
                cropped_image_infos = annotation.crop_images(padding_ratio=self.padding_ratio)
                for index, cropped_image_info in enumerate(cropped_image_infos):
                    if cropped_image_info.geometric_object.content == "":
                        continue

                    if self.rel_image_output_dir is None:
                        image_output_dir = os.path.join(
                            self.data_root_dir,
                            os.path.basename(self.output_file_path).split(".")[0],
                        )
                    else:
                        image_output_dir = os.path.join(
                            self.data_root_dir,
                            self.rel_image_output_dir,
                        )

                    cropped_image_path = TextCropAnnotationWriter.__create_cropped_image_filepath(
                        image_path=annotation.image_path,
                        content=cropped_image_info.geometric_object.content,
                        class_name=cropped_image_info.geometric_object.class_name,
                        index=index,
                        image_output_dir=image_output_dir,
                    )

                    os.makedirs(os.path.dirname(cropped_image_path), exist_ok=True)

                    encoding_safe_imwrite(
                        filename=cropped_image_path,
                        image=cropped_image_info.image,
                    )

                    annotation_file.write(
                        f"{os.path.relpath(cropped_image_path, self.data_root_dir)}"
                        f"{self.seperator}"
                        f"{cropped_image_info.geometric_object.content}\n"
                    )

        return self.output_file_path



# Copyright Open Logistics Foundation
#
# Licensed under the Open Logistics Foundation License 1.3.
# For details on the licensing terms, see the LICENSE file.
# SPDX-License-Identifier: OLFL-1.3

import logging

from shapely.errors import GEOSException
from shapely.geometry import Polygon

from mlcvzoo_base.api.data.box import GeometricPerception

logger = logging.getLogger(__name__)


def compute_iou(
    object_1: GeometricPerception,
    object_2: GeometricPerception,
) -> float:
    """
    Determine the Intersection-over-Union (IoU) of two objects

    Args:
        object_1: First object to compare
        object_2: Second object to compare

    Returns:
        The IoU between object_1 and object_2
    """

    iou: float

    try:
        gt_polygon = Polygon(object_1.polygon())
        polygon = Polygon(object_2.polygon())
        intersect = gt_polygon.intersection(polygon).area
        if intersect == 0.0:
            iou = 0.0
        else:
            union = gt_polygon.union(polygon).area
            iou = intersect / union
    except (ValueError, GEOSException):
        logger.exception("%s, return 0.0 IoU")
        iou = 0.0

    return iou



# Copyright Open Logistics Foundation
#
# Licensed under the Open Logistics Foundation License 1.3.
# For details on the licensing terms, see the LICENSE file.
# SPDX-License-Identifier: OLFL-1.3

"""Module to provide a registry for model and configuration constructors"""

import copy
import inspect
import logging
import os
from typing import Any, Dict, List, Optional, Type, cast

from attr import define
from config_builder import BaseConfigClass, ConfigBuilder
from related import ChildField, SequenceField

from mlcvzoo_base.api.model import ConfigurationType, DataType, Model, PredictionType
from mlcvzoo_base.api.registry import MLCVZooRegistry, RegistryEntry
from mlcvzoo_base.configuration.config_registry import ConfigRegistry
from mlcvzoo_base.configuration.model_config import ModelConfig
from mlcvzoo_base.configuration.replacement_config import (
    STRING_REPLACEMENT_MAP_KEY,
    ReplacementConfig,
)

logger = logging.getLogger(__name__)


@define
class ModelRegistryEntry(BaseConfigClass):
    model_entry: RegistryEntry = ChildField(cls=RegistryEntry)
    config_entry: RegistryEntry = ChildField(cls=RegistryEntry)


@define
class ModelRegistryConfig(BaseConfigClass):
    registry_entries: List[ModelRegistryEntry] = SequenceField(cls=ModelRegistryEntry)


_default_registry_config: ModelRegistryConfig = ModelRegistryConfig(
    registry_entries=[
        ModelRegistryEntry(
            model_entry=RegistryEntry(
                module_type_name="read_from_file_classification",
                module_constructor="ReadFromFileClassificationModel",
                package_name="mlcvzoo_base.models.read_from_file.model",
            ),
            config_entry=RegistryEntry(
                module_type_name="read_from_file_config",
                module_constructor="ReadFromFileConfig",
                package_name="mlcvzoo_base.models.read_from_file.configuration",
            ),
        ),
        ModelRegistryEntry(
            model_entry=RegistryEntry(
                module_type_name="read_from_file_object_detection",
                module_constructor="ReadFromFileObjectDetectionModel",
                package_name="mlcvzoo_base.models.read_from_file.model",
            ),
            config_entry=RegistryEntry(
                module_type_name="read_from_file_config",
                module_constructor="ReadFromFileConfig",
                package_name="mlcvzoo_base.models.read_from_file.configuration",
            ),
        ),
        ModelRegistryEntry(
            model_entry=RegistryEntry(
                module_type_name="read_from_file_segmentation",
                module_constructor="ReadFromFileSegmentationModel",
                package_name="mlcvzoo_base.models.read_from_file.model",
            ),
            config_entry=RegistryEntry(
                module_type_name="read_from_file_config",
                module_constructor="ReadFromFileConfig",
                package_name="mlcvzoo_base.models.read_from_file.configuration",
            ),
        ),
        ModelRegistryEntry(
            model_entry=RegistryEntry(
                module_type_name="yolox",
                module_constructor="YOLOXModel",
                package_name="mlcvzoo_yolox.model",
            ),
            config_entry=RegistryEntry(
                module_type_name="yolox_config",
                module_constructor="YOLOXConfig",
                package_name="mlcvzoo_yolox.configuration",
            ),
        ),
        ModelRegistryEntry(
            model_entry=RegistryEntry(
                module_type_name="darknet_object_detection",
                module_constructor="DarknetDetectionModel",
                package_name="mlcvzoo_darknet.model",
            ),
            config_entry=RegistryEntry(
                module_type_name="darknet_config",
                module_constructor="DarknetConfig",
                package_name="mlcvzoo_darknet.configuration",
            ),
        ),
        ModelRegistryEntry(
            model_entry=RegistryEntry(
                module_type_name="mmdetection_object_detection",
                module_constructor="MMObjectDetectionModel",
                package_name="mlcvzoo_mmdetection.object_detection_model",
            ),
            config_entry=RegistryEntry(
                module_type_name="mmdet_config",
                module_constructor="MMDetectionConfig",
                package_name="mlcvzoo_mmdetection.configuration",
            ),
        ),
        ModelRegistryEntry(
            model_entry=RegistryEntry(
                module_type_name="mmdetection_segmentation",
                module_constructor="MMSegmentationModel",
                package_name="mlcvzoo_mmdetection.segmentation_model",
            ),
            config_entry=RegistryEntry(
                module_type_name="mmdet_config",
                module_constructor="MMDetectionConfig",
                package_name="mlcvzoo_mmdetection.configuration",
            ),
        ),
        ModelRegistryEntry(
            model_entry=RegistryEntry(
                module_type_name="mmrotate",
                module_constructor="MMRotateModel",
                package_name="mlcvzoo_mmrotate.rotated_object_detection_model",
            ),
            config_entry=RegistryEntry(
                module_type_name="mmdet_config",
                module_constructor="MMDetectionConfig",
                package_name="mlcvzoo_mmdetection.configuration",
            ),
        ),
        ModelRegistryEntry(
            model_entry=RegistryEntry(
                module_type_name="mmpretrain",
                module_constructor="MMPretrainModel",
                package_name="mlcvzoo_mmpretrain.classification_model",
            ),
            config_entry=RegistryEntry(
                module_type_name="mmpretrain_config",
                module_constructor="MMPretrainConfig",
                package_name="mlcvzoo_mmpretrain.configuration",
            ),
        ),
        ModelRegistryEntry(
            model_entry=RegistryEntry(
                module_type_name="mmocr_text_detection",
                module_constructor="MMOCRTextDetectionModel",
                package_name="mlcvzoo_mmocr.text_detection_model",
            ),
            config_entry=RegistryEntry(
                module_type_name="mmocr_config",
                module_constructor="MMOCRConfig",
                package_name="mlcvzoo_mmocr.configuration",
            ),
        ),
        ModelRegistryEntry(
            model_entry=RegistryEntry(
                module_type_name="mmocr_text_recognition",
                module_constructor="MMOCRTextRecognitionModel",
                package_name="mlcvzoo_mmocr.text_recognition_model",
            ),
            config_entry=RegistryEntry(
                module_type_name="mmocr_config",
                module_constructor="MMOCRConfig",
                package_name="mlcvzoo_mmocr.configuration",
            ),
        ),
        ModelRegistryEntry(
            model_entry=RegistryEntry(
                module_type_name="tf_classification_custom_block",
                module_constructor="CustomBlockModel",
                package_name="mlcvzoo_tf_classification.custom_block.model",
            ),
            config_entry=RegistryEntry(
                module_type_name="tf_classification_custom_block_config",
                module_constructor="CustomBlockConfig",
                package_name="mlcvzoo_tf_classification.custom_block.configuration",
            ),
        ),
        ModelRegistryEntry(
            model_entry=RegistryEntry(
                module_type_name="tf_classification_xception",
                module_constructor="XceptionModel",
                package_name="mlcvzoo_tf_classification.xception.model",
            ),
            config_entry=RegistryEntry(
                module_type_name="tf_classification_xception_config",
                module_constructor="XceptionConfig",
                package_name="mlcvzoo_tf_classification.xception.configuration",
            ),
        ),
    ]
)


class ModelRegistry(MLCVZooRegistry[Type[Model[PredictionType, ConfigurationType, DataType]]]):
    """
    Class to provide a registry for model and configuration constructors
    """

    def __init__(self, configuration_path: Optional[str] = None) -> None:
        MLCVZooRegistry.__init__(self)

        if configuration_path is None:
            configuration_path = os.getenv("MLCVZOO_MODEL_REGISTRY_PATH", None)

        registry_config: ModelRegistryConfig
        if configuration_path is not None and os.path.isfile(configuration_path):
            registry_config = cast(
                ModelRegistryConfig,
                ConfigBuilder(
                    class_type=ModelRegistryConfig, yaml_config_path=configuration_path
                ).configuration,
            )
        else:
            registry_config = _default_registry_config

        self._config_registry = ConfigRegistry()

        self._model_to_config_dict: Dict[str, str] = {}

        for registry_entry in registry_config.registry_entries:
            registered_model_type_name = self.register_external_module(
                module_type_name=registry_entry.model_entry.module_type_name,
                module_constructor=registry_entry.model_entry.module_constructor,
                package_name=registry_entry.model_entry.package_name,
                force=True,
            )
            registered_config_type_name = self._config_registry.register_external_module(
                module_type_name=registry_entry.config_entry.module_type_name,
                module_constructor=registry_entry.config_entry.module_constructor,
                package_name=registry_entry.config_entry.package_name,
                force=True,
            )
            if registered_model_type_name and registered_config_type_name:
                self._model_to_config_dict[registry_entry.model_entry.module_type_name] = (
                    registry_entry.config_entry.module_type_name
                )

    @property
    def model_registry(
        self,
    ) -> Dict[str, Type[Model[PredictionType, ConfigurationType, DataType]]]:
        return self._registry

    @property
    def config_registry(self) -> Dict[str, Type[BaseConfigClass]]:
        return self._config_registry._registry

    def determine_config_class(self, model_type_name: str) -> Type[BaseConfigClass]:
        return self.config_registry[self._model_to_config_dict[model_type_name]]

    def determine_config_class_name(self, model_type_name: str) -> str:
        return self._model_to_config_dict[model_type_name]

    def register_external_model(
        self, model_type_name: str, model_constructor: str, package_name: str
    ) -> None:
        """
        Register an external model
        Args:
            model_type_name: name of the model to register
            model_constructor: name of the constructor of the model to register
            package_name: the full package to import to call the constructor

        Returns:
            Nothing
        """
        self.register_external_module(
            module_type_name=model_type_name,
            module_constructor=model_constructor,
            package_name=package_name,
        )

    def get_registered_models(
        self,
    ) -> Dict[str, Type[Model[PredictionType, ConfigurationType, DataType]]]:
        return copy.deepcopy(self._registry)

    def get_registered_model_type_names(
        self,
    ) -> List[str]:
        return copy.deepcopy(list(self._registry.keys()))

    def get_model_type(
        self, class_type: str
    ) -> Optional[Type[Model[PredictionType, ConfigurationType, DataType]]]:
        if class_type in self._registry:
            return self._registry[class_type]

        return None

    def register_model(
        self, model_type_name: str, model_constructor: Any, force: bool = False
    ) -> None:
        self.register_module(
            module_type_name=model_type_name, module_constructor=model_constructor, force=force
        )

    def build_model_config(
        self,
        model_type_name: str,
        model_configuration_dict: Dict[str, Any],
        string_replacement_map: Optional[Dict[str, str]] = None,
    ) -> ModelConfig:
        return ModelConfig(
            class_type=model_type_name,
            constructor_parameters={
                "configuration": ConfigBuilder.build_configuration_from_dict(
                    config_dict=model_configuration_dict,
                    class_type=self.determine_config_class(model_type_name=model_type_name),
                    string_replacement_map=string_replacement_map,
                ),
            },
        )

    def init_model(
        self,
        model_config: ModelConfig,
        string_replacement_map: Optional[Dict[str, str]] = None,
    ) -> Model[PredictionType, ConfigurationType, DataType]:
        """
        Generic method for instantiating any model that is registered in the model-registry

        Args:
            model_config: The model configuration defining which model should be initialized
            string_replacement_map: (Optional) A dictionary that defines placeholders which can
                                    be used while parsing a configuration file. They can be
                                    understood as variables that can be used to define configs
                                    that are valid across multiple devices.

                                    If no string_replacement_map a default map based on the
                                    ReplacementConfig will be created and used. This allows
                                    to use the attributes of the ReplacementConfig to be
                                    replaced by os environment variables.

        Returns:
            The created model instance
        """

        model_type = self.get_model_type(class_type=model_config.class_type)
        if model_type is not None:
            model: Model[PredictionType, ConfigurationType, DataType]

            init_params: List[Any] = list(inspect.getfullargspec(model_type.__init__).args)
            # We don't need self as parameter in the configuration
            init_params.remove("self")

            if (
                STRING_REPLACEMENT_MAP_KEY not in model_config.constructor_parameters
                and STRING_REPLACEMENT_MAP_KEY in init_params
            ):
                if string_replacement_map is None:
                    string_replacement_map = ReplacementConfig().to_dict()
                model_config.constructor_parameters[STRING_REPLACEMENT_MAP_KEY] = (
                    string_replacement_map
                )

            try:
                model = model_type(**model_config.constructor_parameters)  # type: ignore[arg-type]
            except TypeError as e:
                logger.error(
                    "Please provide the parameters %s, as specified for %s",
                    init_params,
                    model_type,
                )
                raise e
        else:
            message = (
                f"The model '{model_config.class_type}' is not registered! \n"
                f"The registered models are: {self._registry.keys()}"
            )

            logger.error(message)
            raise ValueError(message)

        return model



# Copyright Open Logistics Foundation
#
# Licensed under the Open Logistics Foundation License 1.3.
# For details on the licensing terms, see the LICENSE file.
# SPDX-License-Identifier: OLFL-1.3

"""Dataclass that holds attributes of Annotation objects"""

from typing import Any, Dict, Optional


class AnnotationAttributes:
    """Class for describing different annotation attributes"""

    def __init__(
        self,
        difficult: bool = False,
        occluded: bool = False,
        content: str = "",
        background: bool = False,
        meta_attributes: Optional[Dict[Any, Any]] = None,
    ):
        self.__difficult: bool = difficult
        self.__occluded: bool = occluded
        self.__content: str = content
        self.__background: bool = background
        self.__meta_attributes: Dict[Any, Any] = (
            meta_attributes if meta_attributes is not None else {}
        )

    @property
    def difficult(self) -> bool:
        """Attribute that defines that an object annotation should be handled as
        "difficult" to see in the image. It can be used later on in the AnnotationHandler
        to filter out specific annotations.

        Returns:
            Whether the object should be handled as "difficult"
        """
        return self.__difficult

    @property
    def occluded(self) -> bool:
        """Attribute that defines that an annotation object is occluded by another
        object in the image. It can be used later on in the AnnotationHandler to filter
        out specific annotations.

        Returns:
            Whether the object should be handled as "occluded"
        """
        return self.__occluded

    @property
    def content(self) -> str:
        """Attribute that defines the textual content of an annotation object.

        Returns:
            The content of an annotation object
        """
        return self.__content

    @content.setter
    def content(self, value: str) -> None:
        """Update set content attribute with the given value"""
        self.__content = value

    @property
    def background(self) -> bool:
        """Attribute that defines that an object should be handled as "background".
        Those are objects which are not in the main focus for the training, because
        they are not positioned in the front of the image view. It is mainly used to
        describe the characteristic of an object with more than just "difficult"
        or occluded. It can be used later on in the annotation handler
        to filter out specific annotations.

        Returns:
            Whether the object should be handled as "background"
        """
        return self.__background

    @property
    def meta_attributes(self) -> Dict[Any, Any]:
        """Attribute that is utilized to store any meta attribute information that
        should be associated with an annotation object.

        Returns:
            The meta attributes of an annotation object as dictionary
        """
        return self.__meta_attributes



# Copyright Open Logistics Foundation
#
# Licensed under the Open Logistics Foundation License 1.3.
# For details on the licensing terms, see the LICENSE file.
# SPDX-License-Identifier: OLFL-1.3

"""Module that holds classes for different types of basic dataset information."""
import os
from dataclasses import dataclass, field
from typing import Dict

from dataclasses_json import dataclass_json

from mlcvzoo_base.api.data.annotation import BaseAnnotation
from mlcvzoo_base.utils.implicit_path_replacements import ImplicitReplacement


@dataclass_json
@dataclass
class ClassesCountInfo:
    count: str
    count_s: str
    count_m: str
    count_l: str


@dataclass_json
@dataclass
class DatasetDirectoryInfo:
    """Class holds information about a specific directory"""

    file_count: int
    classes_count_info: Dict[str, ClassesCountInfo] = field(default_factory=dict)


@dataclass_json
@dataclass
class BaseDatasetInfo:
    """Class holds information about a specific dataset"""

    classes_path: str
    image_count: float

    directory_info: Dict[str, DatasetDirectoryInfo] = field(default_factory=dict)

    base_file: str = ""

    def update(self, annotation: BaseAnnotation) -> None:
        """
        Updates the class's directory information with the given annotation.

        Args:
            annotation: BaseAnnotation object

        """

        self.image_count += 1

        # get relative directory of the annotation
        relative_dirname = os.path.dirname(
            ImplicitReplacement.replace_directory_in_path(
                file_path=annotation.annotation_path,
                replacement_key=annotation.annotation_dir,
                replacement_value="",
            )
        )

        # Init info dict
        if relative_dirname not in self.directory_info:
            self.directory_info[relative_dirname] = DatasetDirectoryInfo(file_count=1)
        else:
            self.directory_info[relative_dirname].file_count += 1

        # Fill info dict with statistics about the annotation
        # TODO:
        # for bounding_box in annotation.get_bounding_boxes(include_segmentations=True):
        #     bbox_height = (bounding_box.box.ymax - bounding_box.box.ymin) / annotation.image_shape[0] * 480
        #     bbox_width = (bounding_box.box.xmax - bounding_box.box.xmin) / annotation.image_shape[1] * 640
        #
        #     bbox_sqrt_area = math.sqrt(bbox_height * bbox_width)
        #
        #     if bounding_box.class_name in info_dict[relative_dirname]:
        #         info_dict[relative_dirname][bounding_box.class_name]["count"] += 1
        #
        #         if bbox_sqrt_area <= 32:
        #             info_dict[relative_dirname][bounding_box.class_name]["count_s"] += 1
        #
        #         if 32 < bbox_sqrt_area <= 96:
        #             info_dict[relative_dirname][bounding_box.class_name]["count_m"] += 1
        #
        #         if bbox_sqrt_area > 96:
        #             info_dict[relative_dirname][bounding_box.class_name]["count_l"] += 1
        #     else:
        #         info_dict[relative_dirname][bounding_box.class_name] = dict()
        #
        #         info_dict[relative_dirname][bounding_box.class_name]["count"] = 1
        #
        #         info_dict[relative_dirname][bounding_box.class_name]["count_s"] = 0
        #         info_dict[relative_dirname][bounding_box.class_name]["count_m"] = 0
        #         info_dict[relative_dirname][bounding_box.class_name]["count_l"] = 0
        #
        #         if bbox_sqrt_area <= 32:
        #             info_dict[relative_dirname][bounding_box.class_name]["count_s"] += 1
        #
        #         if 32 < bbox_sqrt_area <= 96:
        #             info_dict[relative_dirname][bounding_box.class_name]["count_m"] += 1
        #
        #         if bbox_sqrt_area > 96:
        #             info_dict[relative_dirname][bounding_box.class_name]["count_l"] += 1



# Copyright Open Logistics Foundation
#
# Licensed under the Open Logistics Foundation License 1.3.
# For details on the licensing terms, see the LICENSE file.
# SPDX-License-Identifier: OLFL-1.3

""" Class for Segmentation (polygon areas) annotations"""
from __future__ import annotations

import math
from typing import Any, Dict, Optional, Tuple, Union

import numpy as np

from mlcvzoo_base.api.data.annotation_attributes import AnnotationAttributes
from mlcvzoo_base.api.data.bounding_box import BoundingBox
from mlcvzoo_base.api.data.box import Box
from mlcvzoo_base.api.data.class_identifier import ClassIdentifier
from mlcvzoo_base.api.data.classification import Classification
from mlcvzoo_base.api.data.geometric_classifiction import GeometricClassification
from mlcvzoo_base.api.data.types import PolygonType, PolygonTypeNP, polygon_equal
from mlcvzoo_base.api.structs import float_equality_precision


class Segmentation(GeometricClassification, AnnotationAttributes):
    """
    A class for defining the data object consumed by Segmentation models.
    It is mainly described by its polygon (list of 2D points) attribute.
    The polygon captures an area of the image, that is defined by the
    inner of the linear connected points of the polygon and is associated
    with a certain class.
    In addition, a segmentation CAN have an attribute "box" that defines the rectangle
    which encloses the area of the polygon.
    """

    def __init__(
        self,
        polygon: Union[PolygonType, PolygonTypeNP],
        class_identifier: ClassIdentifier,
        score: float,
        difficult: bool = False,
        occluded: bool = False,
        content: str = "",
        model_class_identifier: Optional[ClassIdentifier] = None,
        background: bool = False,
        meta_attributes: Optional[Dict[Any, Any]] = None,
    ):
        Classification.__init__(
            self,
            class_identifier=class_identifier,
            model_class_identifier=model_class_identifier,
            score=score,
        )
        AnnotationAttributes.__init__(
            self,
            difficult=difficult,
            occluded=occluded,
            background=background,
            content=content,
            meta_attributes=meta_attributes,
        )
        self.__polygon: PolygonTypeNP = np.asarray(polygon)

        # The boxes are created lazily
        self.__box: Optional[Box] = None
        self.__ortho_box: Optional[Box] = None

    def polygon(self) -> PolygonTypeNP:
        return self.__polygon

    def ortho_box(self) -> Box:
        if self.__ortho_box is None:
            self.__ortho_box = Box.from_polygon(polygon=self.polygon(), orthogonal=True)
        return self.__ortho_box

    def box(self) -> Box:
        if self.__box is None:
            self.__box = Box.from_polygon(polygon=self.polygon(), orthogonal=False)
        return self.__box

    def __eq__(self, other: object) -> bool:
        try:
            # NOTE: Since floats may very for different systems, don't check the score for equality,
            #       but allow it to be in a reasonable range
            return (
                self.class_identifier.class_id == other.class_identifier.class_id  # type: ignore[attr-defined]
                and self.class_identifier.class_name == other.class_identifier.class_name  # type: ignore[attr-defined]
                and self.model_class_identifier.class_id
                == other.model_class_identifier.class_id  # type: ignore[attr-defined]
                and self.model_class_identifier.class_name
                == other.model_class_identifier.class_name  # type: ignore[attr-defined]
                and self.occluded == other.occluded  # type: ignore[attr-defined]
                and self.difficult == other.difficult  # type: ignore[attr-defined]
                and self.background == other.background  # type: ignore[attr-defined]
                and self.content == other.content  # type: ignore[attr-defined]
                and self.meta_attributes == other.meta_attributes  # type: ignore[attr-defined]
                and math.isclose(a=self.score, b=other.score, abs_tol=float_equality_precision)  # type: ignore[attr-defined]
                and self.box() == other.box()  # type: ignore[attr-defined]
                and polygon_equal(polygon=self.polygon(), other_polygon=other.polygon())  # type: ignore[attr-defined]
            )
        except AttributeError:
            return False

    def __str__(self):  # type: ignore
        return (
            f"Segmentation("
            f'class_identifier="{str(self.class_identifier)}", '
            f'model_class_identifier="{str(self.model_class_identifier)}", '
            f"polygon={self.__polygon.tolist()}, "
            f"score={self.score}, "
            f"difficult={self.difficult}, "
            f"occluded={self.occluded}, "
            f"background={self.background}, "
            f'content="{self.content}", '
            f"meta_attributes={self.meta_attributes}"
            f")"
        )

    def __repr__(self):  # type: ignore
        return (
            f"Segmentation("
            f"class_identifier={repr(self.class_identifier)}, "
            f"model_class_identifier={repr(self.model_class_identifier)}, "
            f"polygon={self.__polygon.tolist()}, "
            f"score={self.score}, "
            f"difficult={self.difficult}, "
            f"occluded={self.occluded}, "
            f"background={self.background}, "
            f'content="{self.content}", '
            f"meta_attributes={self.meta_attributes}"
            f")"
        )

    def to_dict(self, raw_type: bool = False, reduced: bool = False) -> Dict[str, Any]:
        """
        Generate a dictionary representation of the object. The raw_type defined whether
        all data attributes should be converted to python native data types or should
        be kept as is. The reduced parameter allows to create a more compact representation
        where only the main attributes of the object are present.

        Args:
            raw_type: Whether to convert the data attributes to python native data types
            reduced: Whether to create a more compact representation

        Returns:
            The created dictionary representation
        """
        if reduced:
            return {
                "polygon": self.__polygon if raw_type else self.__polygon.tolist(),
                "class_id": self.class_id,
                "class_name": self.class_name,
                "model_class_id": self.model_class_identifier.class_id,
                "model_class_name": self.model_class_identifier.class_name,
                "score": self.score,
            }

        return {
            "polygon": self.__polygon if raw_type else self.__polygon.tolist(),
            "class_identifier": (
                self.class_identifier if raw_type else self.class_identifier.to_dict()
            ),
            "model_class_identifier": (
                self.model_class_identifier if raw_type else self.model_class_identifier.to_dict()
            ),
            "score": self.score,
            "difficult": self.difficult,
            "occluded": self.occluded,
            "background": self.background,
            "content": self.content,
        }

    @staticmethod
    def from_dict(input_dict: Dict[str, Any], reduced: bool = False) -> Segmentation:
        # fmt: off
        if reduced:
            return Segmentation(**{
                "polygon": input_dict["polygon"],
                "class_identifier": ClassIdentifier(**{
                    "class_id": input_dict["class_id"],
                    "class_name": input_dict["class_name"],
                }),
                "model_class_identifier": ClassIdentifier(**{
                    "class_id": input_dict["model_class_id"],
                    "class_name": input_dict["model_class_name"],
                }),
                "score": input_dict["score"],
                "difficult": False,
                "occluded": False,
                "background": False,
                "content": "",
            })
        else:
            return Segmentation(**{
                "polygon": input_dict["polygon"],
                "class_identifier": ClassIdentifier(
                    **input_dict["class_identifier"]
                ),
                "model_class_identifier": ClassIdentifier(
                    **input_dict["model_class_identifier"]
                ),
                "score": input_dict["score"],
                "difficult": input_dict["difficult"],
                "occluded": input_dict["occluded"],
                "background": input_dict["background"],
                "content": input_dict["content"],
            })
        # fmt: on

    def to_json(self) -> Any:
        return self.to_dict(raw_type=False)

    def to_bounding_box(self, image_shape: Optional[Tuple[int, int]] = None) -> BoundingBox:
        """
        Manually build the bounding box object which covers the rectangular area
        that is defined by the polygon attribute

        Returns:
            A Bounding Box version of this Segmentation instance
        """
        box = self.box()

        if image_shape is not None:
            box.clamp(shape=image_shape)

        return BoundingBox(
            class_identifier=self.class_identifier,
            model_class_identifier=self.model_class_identifier,
            score=self.score,
            box=box,
            difficult=self.difficult,
            occluded=self.occluded,
            background=self.background,
            content=self.content,
        )

    def to_points_string(self) -> str:
        """
        Returns:
            Sting, concatenation of the coordinates of polygon attribute
        """

        points_string = ""
        for i, point in enumerate(self.__polygon):
            if i < len(self.__polygon) - 1:
                points_string = (
                    points_string + str(float(point[0])) + "," + str(float(point[1])) + ";"
                )
            else:
                points_string = points_string + str(float(point[0])) + "," + str(float(point[1]))

        return points_string

    def to_rect_segmentation(self) -> Segmentation:
        return Segmentation(
            polygon=self.polygon_to_rect_polygon(polygon=self.polygon()),
            class_identifier=self.class_identifier,
            score=self.score,
            difficult=self.difficult,
            occluded=self.occluded,
            background=self.background,
            content=self.content,
            model_class_identifier=self.model_class_identifier,
        )

    def copy_segmentation(self, class_identifier: ClassIdentifier) -> Segmentation:
        return Segmentation(
            polygon=self.__polygon,
            class_identifier=class_identifier,
            score=self.score,
            difficult=self.difficult,
            occluded=self.occluded,
            background=self.background,
            content=self.content,
            model_class_identifier=self.model_class_identifier,
        )



# Copyright Open Logistics Foundation
#
# Licensed under the Open Logistics Foundation License 1.3.
# For details on the licensing terms, see the LICENSE file.
# SPDX-License-Identifier: OLFL-1.3

"""Module to provide a registry for MLCVZoo data classes"""

import logging
from typing import Any, Dict, Optional, Type, Union

from mlcvzoo_base.api.data.bounding_box import BoundingBox
from mlcvzoo_base.api.data.classification import Classification
from mlcvzoo_base.api.data.ocr_perception import OCRPerception
from mlcvzoo_base.api.data.segmentation import Segmentation
from mlcvzoo_base.api.registry import MLCVZooRegistry

logger = logging.getLogger(__name__)


class DataRegistry(MLCVZooRegistry[Type[Classification]]):
    """
    Class to provide a registry for MLCVZoo data classes that is
    handling the encoding and decoding of such classes.
    """

    __type_key__ = "type"
    __data_key__ = "data"

    def __init__(self) -> None:
        MLCVZooRegistry.__init__(self)

        self.__data_types = [Classification, BoundingBox, Segmentation, OCRPerception]

        for d in self.__data_types:
            self.register_module(d.__name__, d)

    def encode(
        self, prediction: Union[Classification, BoundingBox, Segmentation, OCRPerception]
    ) -> Dict[str, Any]:
        """
        Encode the given data object to a dictionary, it has to be in one of
        the following formats: Classification, BoundingBox, Segmentation, OCRPerception.

        Args:
            prediction: The data to encode

        Returns:
            The encoded data
        """
        prediction_type = prediction.__class__.__name__

        if prediction_type in self._registry:
            return {
                DataRegistry.__type_key__: prediction_type,
                DataRegistry.__data_key__: prediction.to_dict(),
            }

        raise ValueError(
            "Type='%s' is not registered in the MLCVZoo DataRegistry", prediction_type
        )

    def decode(
        self, data: Dict[str, Any]
    ) -> Optional[Union[Classification, BoundingBox, Segmentation, OCRPerception]]:
        """
        Decode the data object from a dictionary

        Args:
            data: The data to decode

        Returns:
            The decoded object or None if the dictionary data does not
            contain decodable data
        """

        try:
            if data[DataRegistry.__type_key__] in self._registry:
                return self._registry[data[DataRegistry.__type_key__]].from_dict(
                    data[DataRegistry.__data_key__]
                )
        except KeyError as key_error:
            logger.error("The key '%s' is missing in the data dictionary", key_error)

        return None



# Copyright Open Logistics Foundation
#
# Licensed under the Open Logistics Foundation License 1.3.
# For details on the licensing terms, see the LICENSE file.
# SPDX-License-Identifier: OLFL-1.3

"""Module for runtimes options."""

from typing import Final

float_equality_precision: Final[float] = 1e-5


class Runtime:
    """Class holding all runtimes a model can be converted to."""

    DEFAULT = "DEFAULT"
    ONNXRUNTIME = "ONNXRUNTIME"
    ONNXRUNTIME_FLOAT16 = "ONNXRUNTIME_FLOAT16"
    TENSORRT = "TENSORRT"
    TENSORRT_INT8 = "TENSORRT_INT8"



# Copyright Open Logistics Foundation
#
# Licensed under the Open Logistics Foundation License 1.3.
# For details on the licensing terms, see the LICENSE file.
# SPDX-License-Identifier: OLFL-1.3

""" Class for Bounding Box information"""

from __future__ import annotations

import logging
import math
from abc import ABC, abstractmethod
from typing import Any, Dict, List, Optional, Tuple, Union

import cv2
import numpy as np

from mlcvzoo_base.api.data.types import (
    FrameShape,
    ImageType,
    Point2D,
    Point2f,
    PolygonType,
    PolygonTypeNP,
    point_equal,
)
from mlcvzoo_base.api.interfaces import Perception
from mlcvzoo_base.configuration.structs import ObjectDetectionBBoxFormats
from mlcvzoo_base.third_party.imutils.perspective import order_points

logger = logging.getLogger(__name__)


class GeometricPerception(ABC, Perception):
    @abstractmethod
    def box(self) -> Box:
        raise NotImplementedError(f"Call of .box() from abstract class {self.__class__}")

    @abstractmethod
    def ortho_box(self) -> Box:
        raise NotImplementedError(f"Call of .ortho_box() from abstract class {self.__class__}")

    @abstractmethod
    def polygon(self) -> PolygonTypeNP:
        raise NotImplementedError(f"Call of .polygon() from abstract class {self.__class__}")

    @staticmethod
    def polygon_to_rect_polygon(polygon: Union[PolygonType, PolygonTypeNP]) -> PolygonTypeNP:
        """
        Returns:
            The minimum orthogonal rectangular area around the polygon of this Segmentation
            represented as 4x2 numpy polygon.
        """

        return order_points(
            polygon=cv2.boxPoints(cv2.minAreaRect(np.asarray(polygon, dtype=np.int_))),
            sort_by_euclidean=False,
        )

    def padded_ortho_box(self, ratio: float, frame_shape: Optional[FrameShape] = None) -> Box:
        ortho_box = self.ortho_box()

        pad_w = int(ortho_box.width * ratio)
        pad_h = int(ortho_box.height * ratio)

        if frame_shape:
            xmax = min(frame_shape.width, ortho_box.xmax + pad_w)
            ymax = min(frame_shape.height, ortho_box.ymax + pad_h)
        else:
            xmax = ortho_box.xmax + pad_w
            ymax = ortho_box.ymax + pad_h

        return Box(
            xmin=max(0, ortho_box.xmin - pad_w),
            ymin=max(0, ortho_box.ymin - pad_h),
            xmax=xmax,
            ymax=ymax,
        )

    def crop_img_by_ratio(self, image: ImageType, padding_ratio: float) -> ImageType:
        """
        Crop image based on the orthogonal box of this object.
        The padding ratio will increase the size of the box by the given
        factor

        Args:
            image: The image to crop from
            padding_ratio: The ratio for increasing the box size

        Returns:
            The cropped image
        """
        padded_box = self.padded_ortho_box(
            ratio=padding_ratio, frame_shape=FrameShape(*image.shape)
        )

        return image[padded_box.ymin : padded_box.ymax, padded_box.xmin : padded_box.xmax]

    def crop_img(
        self, frame: ImageType, margin_x: float = 0.0, margin_y: float = 0.0
    ) -> Optional[ImageType]:
        """
        Create a crop of the given frame based on the information of this box object
        and the given margins. The margin are used as scale factors based on the
        width (x direction) and height (y-direction)

        Args:
            frame: The frame to crop from
            margin_x: The margin around the box in x direction
            margin_y: The margin around the box in y direction

        Returns:
            The cropped image (if it could be computed)
        """

        if frame is None:
            return None

        box = self.ortho_box()

        xmin = box.xmin - round(margin_x * box.width)
        xmax = box.xmax + round(margin_x * box.width)
        ymin = box.ymin - round(margin_y * box.height)
        ymax = box.ymax + round(margin_y * box.height)

        margin_box = Box(xmin=xmin, ymin=ymin, xmax=xmax, ymax=ymax)
        margin_box.clamp(shape=(frame.shape[0], frame.shape[1]))

        cropped_image = frame[
            int(margin_box.ymin) : int(margin_box.ymax),
            int(margin_box.xmin) : int(margin_box.xmax),
        ]

        return cropped_image


class Box(GeometricPerception):
    """
    Class for storing bounding box information.

    The constructor is defined by the parameters upper left point and
    lower right point of the Box, as well as its rotation angle around the center
    of the Box. The points are defined as:
    - Upper left point given by xmin and ymin
    - Lower right point given by xmax and xmax
    |-----------------------|
    |(xmin, ymin)           |
    |                       |
    |                       |
    |          x angle      |
    |                       |
    |                       |
    |           (xmax, ymax)|
    |-----------------------|

    To avoid confusion a modulo operator is applied to the angle, which limits
    the angle value between 0 and 360 degrees.

    The width is determined between the top-left and top-right point.
    The height is determined between the top-left and bottom-left point.

    Internal representation of a Box is defined by its four points:
    - top-left
    - top-right
    - bottom-right
    - bottom-left

      ---       width       ---
      top-left        top-right
      |-----------------------|
      |X                     X|
      |                       |
      |                       |
      |                       |
      |                       |
      |                       |
      |X                     X|
      |-----------------------|
      bottom-left  bottom-right

      REMARK: The attributes xmin, ymin, xmax and ymax are not changed by the rotation and
            always reflect the original coordinates from the orthogonal box. For accessing
            the rotation points, please use top-left, top-right, bottom-right and bottom-left
    """

    def __init__(
        self, xmin: float, ymin: float, xmax: float, ymax: float, angle: float = 0.0
    ) -> None:
        """Creates a Box object

        Args:
            xmin: Upper left x coordinate of the Box
            ymin: Upper left y coordinate of the Box
            xmax: Lower right x coordinate of the Box
            ymax: Lower right y coordinate of the Box
            angle: Rotation angle in degrees around the
                   center point of the Box.

        Returns:
            The created object
        """
        if xmin >= xmax:
            raise ValueError(f"xmin={xmin} has to be < xmax={xmax}")

        if ymin >= ymax:
            raise ValueError(f"ymin={ymin} has to be < ymax={ymax}")

        self._angle: float = angle % 360

        self._xmin: float = float(xmin)
        self._ymin: float = float(ymin)
        self._xmax: float = float(xmax)
        self._ymax: float = float(ymax)

        self._point_0 = [self._xmin, self._ymin]
        self._point_1 = [self._xmax, self._ymin]
        self._point_2 = [self._xmax, self._ymax]
        self._point_3 = [self._xmin, self._ymax]

        self._center: Point2f = self.__center()

        self.__update()

    def __eq__(self, other: object) -> bool:
        try:
            return (
                point_equal(self._point_0, other._point_0)  # type: ignore[attr-defined]
                and point_equal(self._point_1, other._point_1)  # type: ignore[attr-defined]
                and point_equal(self._point_2, other._point_2)  # type: ignore[attr-defined]
                and point_equal(self._point_3, other._point_3)  # type: ignore[attr-defined]
            )
        except AttributeError:
            return False

    def __repr__(self) -> str:
        return (
            f"Box("
            f"xmin={self._xmin}, "
            f"ymin={self._ymin}, "
            f"xmax={self._xmax}, "
            f"ymax={self._ymax}, "
            f"angle={self._angle}"
            f")"
        )

    def to_dict(self) -> Dict[str, float]:
        return {
            "xmin": self._xmin,
            "ymin": self._ymin,
            "xmax": self._xmax,
            "ymax": self._ymax,
            "angle": self._angle,
        }

    def to_json(self) -> Any:
        return self.to_dict()

    def __update(self) -> None:
        if self._angle != 0.0:
            four_point_box = cv2.boxPoints(
                (self._center, (self._xmax - self._xmin, self._ymax - self._ymin), self._angle)
            )
            four_point_box = order_points(polygon=four_point_box, sort_by_euclidean=False)
            # Cast to convert numpy to python int
            self._point_0 = [float(four_point_box[0][0]), float(four_point_box[0][1])]
            self._point_1 = [float(four_point_box[1][0]), float(four_point_box[1][1])]
            self._point_2 = [float(four_point_box[2][0]), float(four_point_box[2][1])]
            self._point_3 = [float(four_point_box[3][0]), float(four_point_box[3][1])]
        else:
            self._point_0 = [self._xmin, self._ymin]
            self._point_1 = [self._xmax, self._ymin]
            self._point_2 = [self._xmax, self._ymax]
            self._point_3 = [self._xmin, self._ymax]

        self._center = self.__center()

    @property
    def xmin(self) -> int:
        return int(self._xmin)

    @property
    def ymin(self) -> int:
        return int(self._ymin)

    @property
    def xmax(self) -> int:
        return int(self._xmax)

    @property
    def ymax(self) -> int:
        return int(self._ymax)

    @property
    def xminf(self) -> float:
        return self._xmin

    @property
    def yminf(self) -> float:
        return self._ymin

    @property
    def xmaxf(self) -> float:
        return self._xmax

    @property
    def ymaxf(self) -> float:
        return self._ymax

    @property
    def angle(self) -> float:
        return self._angle

    @property
    def width(self) -> int:
        return int(self.widthf)

    @property
    def height(self) -> int:
        return int(self.heightf)

    @property
    def widthf(self) -> float:
        if self.is_orthogonal():
            return abs(self.top_left[0] - self.top_right[0])

        return float(
            math.sqrt(
                math.pow(self.top_left[0] - self.top_right[0], 2)
                + math.pow(self.top_left[1] - self.top_right[1], 2)
            )
        )

    @property
    def heightf(self) -> float:
        if self.is_orthogonal():
            return abs(self.top_left[1] - self.bottom_left[1])

        return float(
            math.sqrt(
                math.pow(self.top_left[0] - self.bottom_left[0], 2)
                + math.pow(self.top_left[1] - self.bottom_left[1], 2)
            )
        )

    @property
    def top_left(self) -> Point2f:
        return self._point_0

    @property
    def top_right(self) -> Point2f:
        return self._point_1

    @property
    def bottom_right(self) -> Point2f:
        return self._point_2

    @property
    def bottom_left(self) -> Point2f:
        return self._point_3

    @property
    def top_left_2d(self) -> Point2D:
        return [int(self._point_0[0]), int(self._point_0[1])]

    @property
    def top_right_2d(self) -> Point2D:
        return [int(self._point_1[0]), int(self._point_1[1])]

    @property
    def bottom_right_2d(self) -> Point2D:
        return [int(self._point_2[0]), int(self._point_2[1])]

    @property
    def bottom_left_2d(self) -> Point2D:
        return [int(self._point_3[0]), int(self._point_3[1])]

    def is_orthogonal(self) -> bool:
        return self._angle % 90 == 0.0

    @staticmethod
    def init_format_based(
        box_format: str,
        box_list: Tuple[float, float, float, float],
        src_shape: Optional[Tuple[int, int]] = None,
        dst_shape: Optional[Tuple[int, int]] = None,
        angle: float = 0.0,
    ) -> Box:
        """
        Construct a Box object using the given format and box coordinates

        Args:
            box_format: Specify the way for parsing the box argument
            box_list: The box coordinates as 4D tuple
            src_shape: shape of the original image as tuple (height, width)
            dst_shape: desired shape for creating the bounding boxes as tuple (height, width)
            angle: The rotation angle around the center of the box in degrees

        Returns:
            The constructed Box object
        """

        xmin = max(0.0, box_list[0])
        ymin = max(0.0, box_list[1])

        width = box_list[2]
        if width < 0:
            raise ValueError("Can not build a box with negative width")

        height = box_list[3]
        if height < 0:
            raise ValueError("Can not build a box with negative height")

        if box_format == ObjectDetectionBBoxFormats.XYWH:
            base_box = Box(
                xmin=xmin,
                ymin=ymin,
                xmax=xmin + width,
                ymax=ymin + height,
                angle=angle,
            )
        elif box_format == ObjectDetectionBBoxFormats.XYXY:
            base_box = Box(xmin=xmin, ymin=ymin, xmax=width, ymax=height, angle=angle)
        elif box_format == ObjectDetectionBBoxFormats.CXCYWH:
            base_box = Box(
                xmin=(xmin - width / 2),
                ymin=(ymin - height / 2),
                xmax=(xmin + width / 2),
                ymax=(ymin + height / 2),
                angle=angle,
            )
        else:
            valid_formats = ObjectDetectionBBoxFormats.get_values_as_list(
                class_type=ObjectDetectionBBoxFormats
            )
            raise ValueError(
                f"Format {box_format} is not supported. Please provide any of {valid_formats}"
            )

        if src_shape is not None:
            base_box.clamp(shape=src_shape)

        if src_shape is not None and dst_shape is not None:
            base_box.scale(src_shape=src_shape, dst_shape=dst_shape)

        return base_box

    @staticmethod
    def from_polygon(polygon: Union[PolygonType, PolygonTypeNP], orthogonal: bool = True) -> Box:
        """
        Produce a box object that covers the minimal rectangular area which is
        covered by the given polygon. When the parameter 'orthogonal' is set to true,
        the box will be orthogonal, otherwise a rotated box will be created.

        Args:
            polygon: PolygonType, a list of points that form the polygon
            orthogonal: Whether the created box should be orthogonal or can be rotated

        Returns:
            A Box that holds the coordinates of the bounding rectangle of the polygon
        """

        if orthogonal:
            x, y, w, h = cv2.boundingRect(np.asarray(polygon, dtype=np.int_))

            return Box.init_format_based(
                box_list=(
                    float(x),
                    float(y),
                    float(w),
                    float(h),
                ),
                box_format=ObjectDetectionBBoxFormats.XYWH,
            )

        (x, y), (w, h), angle = cv2.minAreaRect(np.asarray(polygon, dtype=np.int_))

        return Box.init_format_based(
            box_format=ObjectDetectionBBoxFormats.CXCYWH,
            box_list=(int(x), int(y), int(w), int(h)),
            angle=angle,
        )

    def to_list(self, dst_type: Any = int) -> List[Any]:
        """
        Args:
            dst_type: destination type to transform the data to

        Returns:
            List of transformed polygons
        """

        return [
            dst_type(self._xmin),
            dst_type(self._ymin),
            dst_type(self._xmax),
            dst_type(self._ymax),
        ]

    def box(self) -> Box:
        """Creates a copy of this Box object

        Returns:
            The created Box object
        """
        return Box(
            xmin=self._xmin, ymin=self._ymin, xmax=self._xmax, ymax=self._ymax, angle=self._angle
        )

    def ortho_box(self) -> Box:
        """The orthogonal box around this Box object. In case of an already orthogonal
        Box, the same instance is returned, otherwise a (orthogonal) box around the polygon of
        the Box object is created.

        Returns:
            The orthogonal Box object
        """
        if self.is_orthogonal():
            return self.box()

        return Box.from_polygon(polygon=self.polygon(), orthogonal=True)

    def polygon(self) -> PolygonTypeNP:
        return np.asarray(
            [
                self.top_left,
                self.top_right,
                self.bottom_right,
                self.bottom_left,
            ]
        )

    def clamp(self, shape: Tuple[int, int]) -> None:
        """
        Clamps the bounding-box based on the given shape

        Does only work for orthogonal boxes.

        Args:
            shape: The shape to define the min and max coordinates
                   for the clamping, format in (y, x)

        Returns:
            None
        """

        self._xmin = max(0.0, self._xmin)
        self._ymin = max(0.0, self._ymin)
        self._ymax = min(float(shape[0] - 1), self._ymax)
        self._xmax = min(float(shape[1] - 1), self._xmax)

        self.__update()

    def scale(self, src_shape: Tuple[int, int], dst_shape: Tuple[int, int]) -> None:
        """
        Scale the Box according to the given shapes of the source and destination image

        Args:
            src_shape: shape of the original image as tuple (height, width)
            dst_shape: desired shape for creating the bounding boxes as tuple (height, width)

        Returns:
            None
        """
        if src_shape[0] < 0 or src_shape[1] < 0 or src_shape[0] > src_shape[1]:
            raise ValueError("Invalid source shape %s: ", src_shape)

        if dst_shape[0] < 0 or dst_shape[1] < 0 or dst_shape[0] > dst_shape[1]:
            raise ValueError("Invalid destination shape %s: ", dst_shape)

        height_scale_factor = dst_shape[0] / src_shape[0]
        width_scale_factor = dst_shape[1] / src_shape[1]

        self._xmin = self.xmin * width_scale_factor
        self._xmax = self.xmax * width_scale_factor
        self._ymin = self.ymin * height_scale_factor
        self._ymax = self.ymax * height_scale_factor

        self.__update()

    def center(self) -> Point2f:
        """
        Calculates the geometric center of the Box

        Returns:
            A Tuple as the coordinates of the center
        """
        return self._center

    def __center(self) -> Point2f:
        """
        Calculates the center coordinates of the Box

        Returns:
            A Tuple as the coordinates of the center
        """
        return [
            (self._point_0[0] + self._point_2[0]) * 0.5,
            (self._point_0[1] + self._point_2[1]) * 0.5,
        ]

    def translation(self, x: float, y: float) -> None:
        """
        Shifts Box for x and y pixels in x and y direction respectively

        Args:
            x: int value for shift in x direction
            y: int value for shift in y direction

        Returns:
            None
        """
        self._xmin += x
        self._xmax += x
        self._ymin += y
        self._ymax += y

        self.__update()

    def new_center(self, x: float, y: float) -> None:
        """
        Shifts the Box based on a new center coordinate. Scale of Box is kept.

        Args:
            x: int value, x coordinate of the new center
            y: int value, y coordinate of the new center

        Returns:
            None
        """
        self.translation(x=x - self.center()[0], y=y - self.center()[1])

    def color_hist(
        self,
        frame: ImageType,
        margin_x: float,
        margin_y: float,
    ) -> Optional[ImageType]:
        """
        Calculate the color history the for the Box. First a crop
        of the given image is created based on the box information
        of this object and the given margins. The margin are used
        as scale factors based on the width (x direction) and height
        (y-direction).
        Afterwards the histogram is computed for this crop.

        Args:
            frame: The frame to crop from
            margin_x: The margin around the box in x direction
            margin_y: The margin around the box in y direction

        Returns:
            Color histogram of the box
        """

        cropped_image = self.crop_img(frame=frame, margin_x=margin_x, margin_y=margin_y)

        if cropped_image is None:
            return None

        # Convert image to HSV (Hue, Saturation, Value)
        hsv_image: ImageType = cv2.cvtColor(  # pylint: disable=no-member
            cropped_image, cv2.COLOR_BGR2HSV  # pylint: disable=no-member
        )

        h_bins = 10
        s_bins = 10

        # hue varies from 0 to 179, saturation from 0 to 255
        h_ranges = [0, 180]
        s_ranges = [0, 256]

        # Use the 0-th and 1-st channels
        channels = [0, 1]
        color_hist: ImageType = cv2.calcHist(
            [hsv_image],
            channels,
            None,
            [h_bins, s_bins],
            h_ranges + s_ranges,
            accumulate=False,
        )
        cv2.normalize(color_hist, color_hist, alpha=0, beta=1, norm_type=cv2.NORM_MINMAX)

        return color_hist


def compute_iou(box_1: Box, box_2: Box) -> float:
    """
    Determine the Intersection-over-Union (IoU) of two box objects

    IMPORTANT: This method in designed for orthogonal boxes. Therefore,
               the iou is based on the ortho_box of each box object

    Args:
        box_1: Box object 1
        box_2: Box object 2

    Returns:
        The IoU between box_1 and box_2
    """

    # Ensure to have orthogonal boxes!
    box_1 = box_1.ortho_box()
    box_2 = box_2.ortho_box()

    # Determine the coordinates of the intersection rectangle
    x_left = max(box_1.xmin, box_2.xmin)
    x_right = min(box_1.xmax, box_2.xmax)
    y_bottom = min(box_1.ymax, box_2.ymax)
    y_top = max(box_1.ymin, box_2.ymin)

    # Boxes don't overlap at all
    if x_right < x_left or y_bottom < y_top:
        return 0.0

    # The intersection of two axis-aligned bounding boxes
    # is always an axis-aligned bounding box => A(intersect)
    intersection_area = (x_right - x_left) * (y_bottom - y_top)

    # Compute the area of both boxes => A(box_1) and A(box_2(
    box_1_area = (box_1.xmax - box_1.xmin) * (box_1.ymax - box_1.ymin)
    box_2_area = (box_2.xmax - box_2.xmin) * (box_2.ymax - box_2.ymin)

    # Compute the intersection over union by taking the intersection
    # area and dividing it by the sum of box areas:
    # => iou = A(intersect) / ( A(box_1) + A(box_2))
    iou = intersection_area / float(box_1_area + box_2_area - intersection_area)

    # Clamp the iou value
    return max(min(iou, 1.0), 0.0)


def euclidean_distance(box_1: Box, box_2: Box) -> float:
    """
    Determine the euclidean distance between two Box objects

    Args:
        box_1: Box object 1
        box_2: Box object 2

    Returns:
        The euclidean distance between two Box objects
    """

    center1 = box_1.center()
    center2 = box_2.center()

    return float(((center2[0] - center1[0]) ** 2 + (center2[1] - center1[1]) ** 2) ** 0.5)


def rotate_point(point: Point2f, rotation_origin: Point2f, angle: float) -> Point2f:
    """
    Rotate a given point around an origin point with a given angle

    Args:
        point: The point to rotate
        rotation_origin: The origin point around which the point is rotated
        angle: The rotation angle

    Returns:
        The rotated point
    """
    x0 = rotation_origin[0]
    x1 = point[0]

    y0 = rotation_origin[1]
    y1 = point[1]

    rad_angle = angle * (math.pi / 180)

    x2 = ((x1 - x0) * math.cos(rad_angle)) - ((y1 - y0) * math.sin(rad_angle)) + x0
    y2 = ((x1 - x0) * math.sin(rad_angle)) + ((y1 - y0) * math.cos(rad_angle)) + y0

    return [float(x2), float(y2)]



# Copyright Open Logistics Foundation
#
# Licensed under the Open Logistics Foundation License 1.3.
# For details on the licensing terms, see the LICENSE file.
# SPDX-License-Identifier: OLFL-1.3

"""Module for building Pascal VOC formatted annotations."""
import logging
import os
import xml.etree.ElementTree as ET_xml
from typing import Any, Dict, List, Tuple

from mlcvzoo_base.api.data.annotation import BaseAnnotation
from mlcvzoo_base.api.data.annotation_builder import AnnotationBuilder
from mlcvzoo_base.api.data.annotation_class_mapper import AnnotationClassMapper
from mlcvzoo_base.api.data.bounding_box import BoundingBox
from mlcvzoo_base.api.data.box import Box
from mlcvzoo_base.api.data.class_identifier import ClassIdentifier
from mlcvzoo_base.api.exceptions import ClassMappingNotFoundError
from mlcvzoo_base.configuration.structs import ObjectDetectionBBoxFormats
from mlcvzoo_base.configuration.utils import str2bool
from mlcvzoo_base.data_preparation.utils import ensure_abspath

logger = logging.getLogger(__name__)


class PascalVOCAnnotationBuilder(AnnotationBuilder):
    """
    Super class for defining the methods that are needed to build a BaseAnnotation
    object from a Pascal VOC type XML file.
    """

    def __init__(
        self,
        mapper: AnnotationClassMapper,
        use_difficult: bool = True,
        use_occluded: bool = True,
        use_background: bool = True,
    ) -> None:
        AnnotationBuilder.__init__(self)

        self.mapper = mapper
        self.use_difficult = use_difficult
        self.use_occluded = use_occluded
        self.use_background = use_background

    def build(
        self,
        image_path: str,
        annotation_path: str,
        image_dir: str,
        annotation_dir: str,
        replacement_string: str,
    ) -> BaseAnnotation:
        (
            annotation_path,
            image_shape,
            bounding_boxes,
            source_image_filename,
        ) = self.__init_from_pascal_voc_xml(xml_file_path=annotation_path)

        if not os.path.isfile(image_path):
            image_path = source_image_filename

        annotation: BaseAnnotation = BaseAnnotation(
            image_path=image_path,
            annotation_path=annotation_path,
            image_shape=image_shape,
            classifications=[],
            bounding_boxes=bounding_boxes,
            segmentations=[],
            image_dir=image_dir,
            annotation_dir=annotation_dir,
            replacement_string=replacement_string,
        )

        try:
            AnnotationBuilder._check_and_fix_annotation(annotation=annotation)
        except ValueError as value_error:
            logger.exception(
                f"{value_error}, in a future version, the whole annotation will be skipped!"
            )

        annotation = ensure_abspath(annotation=annotation)

        return annotation

    def __init_from_pascal_voc_xml(
        self,
        xml_file_path: str,
    ) -> Tuple[str, Tuple[int, int], List[BoundingBox], str]:
        """
        Parses and generates objects that are needed to build a BaseAnnotation

        Args:
            xml_file_path: Can either be a string to a XML-File or a File-Object of a XML-File
                in VOC-Format

        Returns: the given xml_file_path,
                the shape of the annotated image,
                a list of annotated bounding_boxes and
                the file path of the image, if available in xml file

        """

        bounding_boxes: List[BoundingBox] = list()

        tree = ET_xml.parse(xml_file_path)

        root = tree.getroot()

        width = int(root.find("size").find("width").text)  # type: ignore
        height = int(root.find("size").find("height").text)  # type: ignore
        image_shape = (height, width)

        source_image_filename = ""
        if root.find("path") is not None:
            source_image_filename = (
                root.find("path").text if root.find("path").text is not None else ""  # type: ignore
            )

        for element in root.findall("object"):
            annotation_class_name: str = element[0].text  # type: ignore

            try:
                # map the parsed "class_name" according to the mapping defined in the mapper class
                class_name = self.mapper.map_annotation_class_name_to_model_class_name(
                    class_name=annotation_class_name
                )

                class_id = self.mapper.map_annotation_class_name_to_model_class_id(
                    class_name=annotation_class_name
                )
            except ClassMappingNotFoundError:
                logger.debug(
                    "Could not find a valid class-mapping for class-name '%s'. "
                    "BndBox will be skipped, file = '%s'",
                    annotation_class_name,
                    xml_file_path,
                )
                continue

            box_tag = element.find("bndbox")

            if box_tag is None:
                continue

            difficult, occluded = self.__parse_default_attributes(element=element)
            background, content, meta_attributes = self.__parse_attributes(element=element)

            try:
                bounding_box = BoundingBox(
                    box=Box.init_format_based(
                        box_list=(
                            float(str(box_tag[0].text)),
                            float(str(box_tag[1].text)),
                            float(str(box_tag[2].text)),
                            float(str(box_tag[3].text)),
                        ),
                        box_format=ObjectDetectionBBoxFormats.XYXY,
                        src_shape=image_shape,
                    ),
                    class_identifier=ClassIdentifier(class_id=class_id, class_name=class_name),
                    model_class_identifier=ClassIdentifier(
                        class_id=class_id, class_name=class_name
                    ),
                    difficult=difficult,
                    background=background,
                    occluded=occluded,
                    content=content,
                    meta_attributes=meta_attributes,
                    score=1.0,
                )
            except ValueError as ve:
                logger.exception("%s, Bounding box will be skipped..." % (ve))
                continue

            if difficult and not self.use_difficult:
                continue

            if occluded and not self.use_occluded:
                continue

            if background and not self.use_background:
                continue

            bounding_boxes.append(bounding_box)

        return xml_file_path, image_shape, bounding_boxes, source_image_filename

    @staticmethod
    def __parse_default_attributes(element: ET_xml.Element) -> Tuple[bool, bool]:
        difficult = False
        difficult_tag = element.find("difficult")
        if difficult_tag is not None and difficult_tag.text is not None:
            difficult = str2bool(difficult_tag.text)

        occluded = False
        occluded_tag = element.find("occluded")
        if occluded_tag is not None and occluded_tag.text is not None:
            occluded = str2bool(occluded_tag.text)

        return difficult, occluded

    @staticmethod
    def __parse_attributes(element: ET_xml.Element) -> Tuple[bool, str, Dict[str, Any]]:
        background: bool = False
        content: str = ""
        meta_attributes: Dict[str, Any] = {}
        attributes_tag = element.find("attributes")
        if attributes_tag is None:
            return background, content, meta_attributes

        for attribute in attributes_tag:
            _attribute_name = attribute.find("name")

            if _attribute_name is not None:
                if _attribute_name.text == "background":
                    background = str2bool(
                        PascalVOCAnnotationBuilder.__parse_attribute_value(
                            attribute_element=attribute, default=False
                        )
                    )
                elif _attribute_name.text == "content":
                    content = PascalVOCAnnotationBuilder.__parse_attribute_value(
                        attribute_element=attribute, default=""
                    )
                elif _attribute_name.text is not None:
                    _attribute_value_element = attribute.find("value")
                    if _attribute_value_element is not None:
                        meta_attributes[_attribute_name.text] = _attribute_value_element.text

        return background, content, meta_attributes

    @staticmethod
    def __parse_attribute_value(attribute_element: ET_xml.Element, default: Any) -> Any:
        _attribute_value_element = attribute_element.find("value")
        if _attribute_value_element is not None:
            _attribute_value = _attribute_value_element.text

            if _attribute_value is not None:
                return _attribute_value

        return default



# Copyright Open Logistics Foundation
#
# Licensed under the Open Logistics Foundation License 1.3.
# For details on the licensing terms, see the LICENSE file.
# SPDX-License-Identifier: OLFL-1.3

import copy
import logging
from enum import Enum
from typing import Dict, List, Optional, Tuple

import numpy as np
from tqdm import tqdm

from mlcvzoo_base.api.data.annotation import BaseAnnotation
from mlcvzoo_base.api.data.annotation_class_mapper import AnnotationClassMapper
from mlcvzoo_base.api.data.class_identifier import ClassIdentifier
from mlcvzoo_base.api.data.geometric_classifiction import GeometricClassification
from mlcvzoo_base.evaluation.geometric.data_classes import (
    CONFUSION_MATRIX_DICT_TYPE,
    DEFAULT_FLOAT_VALUE,
    DEFAULT_INT_VALUE,
    METRIC_DICT_TYPE,
    METRIC_IMAGE_INFO_TYPE,
    EvaluationEntry,
    GeometricEvaluationComputingData,
    GeometricEvaluationMetrics,
    GeometricMetrics,
    MetricImageInfo,
)
from mlcvzoo_base.evaluation.geometric.structs import GeometricSizeTypes
from mlcvzoo_base.evaluation.geometric.utils import (
    compute_max_prediction,
    get_bbox_size_type,
)
from mlcvzoo_base.third_party.py_faster_rcnn.voc_ap import voc_ap

logger = logging.getLogger(__name__)


class EvaluationContexts(str, Enum):
    ALL = "ALL"
    OBJECT_DETECTION = "OBJECT_DETECTION"
    ROTATED_OBJECT_DETECTION = "ROTATED_OBJECT_DETECTION"
    SEGMENTATION = "SEGMENTATION"

    def __repr__(self):  # type: ignore[no-untyped-def]
        return str.__repr__(self)

    def __str__(self):  # type: ignore[no-untyped-def]
        return str.__str__(self)


class MetricsComputation:
    """
    Main class for handling the evaluation of object detection models.
    """

    iou_thresholds_ap_50_95: List[float] = [
        0.5,
        0.55,
        0.6,
        0.65,
        0.7,
        0.75,
        0.8,
        0.85,
        0.9,
        0.95,
    ]

    def __init__(
        self,
        model_specifier: str,
        iou_thresholds: List[float],
        gt_annotations: List[BaseAnnotation],
        predictions_list: List[List[GeometricClassification]],
        mapper: AnnotationClassMapper,
        evaluation_context: str = EvaluationContexts.ALL.value,
    ):
        self.model_specifier = model_specifier
        self.iou_thresholds: List[float] = iou_thresholds
        self.dataset_length: int = len(gt_annotations)

        self.class_identifier_list: List[ClassIdentifier] = []

        self.mapper = mapper
        self.class_identifier_list = self.mapper.create_class_identifier_list()

        # Remove duplicates
        self.class_identifier_list = list(dict.fromkeys(self.class_identifier_list))

        # 1st key: dataset index
        # 2nd key: class-identifier
        self._all_predictions_dict: Dict[int, Dict[str, List[EvaluationEntry]]] = {
            i: {} for i in range(self.dataset_length)
        }
        self._all_gt_dict: Dict[int, Dict[str, List[EvaluationEntry]]] = {
            i: {} for i in range(self.dataset_length)
        }

        self.model_metrics: GeometricEvaluationMetrics = GeometricEvaluationMetrics(
            model_specifier=self.model_specifier
        )
        self.computing_data: GeometricEvaluationComputingData = GeometricEvaluationComputingData()

        process_bar = tqdm(
            zip(gt_annotations, predictions_list),
            desc=f"Compute metrics",
        )

        for index, (gt_annotation, predictions) in enumerate(process_bar):
            self.__update_from_prediction(
                index=index,
                image_path=gt_annotation.image_path,
                gt_objects=MetricsComputation.__filter_gt_objects(
                    evaluation_context=evaluation_context, gt_annotation=gt_annotation
                ),
                predictions=predictions,
            )

    @staticmethod
    def __filter_gt_objects(
        evaluation_context: str, gt_annotation: BaseAnnotation
    ) -> List[GeometricClassification]:
        gt_objects: List[GeometricClassification]
        if evaluation_context == EvaluationContexts.ALL.value:
            gt_objects = gt_annotation.bounding_boxes + gt_annotation.segmentations  # type: ignore[assignment, operator]
        elif evaluation_context == EvaluationContexts.OBJECT_DETECTION.value:
            gt_objects = []
            for bounding_box in gt_annotation.bounding_boxes:
                if bounding_box.box().is_orthogonal():
                    gt_objects.append(bounding_box)
        elif evaluation_context == EvaluationContexts.ROTATED_OBJECT_DETECTION.value:
            gt_objects = gt_annotation.bounding_boxes  # type: ignore[assignment]
        elif evaluation_context == EvaluationContexts.SEGMENTATION.value:
            gt_objects = gt_annotation.segmentations  # type: ignore[assignment]
        else:
            raise ValueError(
                "EvaluationContext '%s' not valid, must be one of %s"
                % (
                    evaluation_context,
                    [EvaluationContexts.OBJECT_DETECTION, EvaluationContexts.SEGMENTATION],
                )
            )

        return gt_objects

    def get_metrics_dict(self) -> METRIC_DICT_TYPE:
        return self.model_metrics.metrics_dict

    def get_metrics_image_info_dict(
        self,
    ) -> METRIC_IMAGE_INFO_TYPE:
        return self.model_metrics.metrics_image_info_dict

    @staticmethod
    def get_overall_ap(metrics_dict: METRIC_DICT_TYPE, iou_threshold: float) -> float:
        """
        Calculate AP metric over every class specific AP metric for bounding boxes of all sizes.

        Args:
            metrics_dict: The dictionary that stores the relevant metric data, which is the
                          basis for the calculation
            iou_threshold: The iou threshold for which the AP metric should be computed

        Returns:
            The computed overall AP metric
        """

        if iou_threshold in metrics_dict:
            class_metrics = metrics_dict[iou_threshold][GeometricSizeTypes.BBOX_ALL]
        else:
            raise ValueError(
                "Can not compute overall AP for iou-threshold=%s, "
                "no data is given in the metrics-dict.",
                iou_threshold,
            )

        ap_list: List[float] = [class_metric.AP for class_metric in class_metrics.values()]
        return sum(ap_list) / len(ap_list)

    @staticmethod
    def compute_average_ap(model_metrics: GeometricEvaluationMetrics) -> float:
        """
        Compute the average of the AP metric for every
        overall AP (average over all classes) of all iou-thresholds
        in the metrics_dict.

        Args:
            model_metrics: The model-metrics for which to compute the average ap

        Returns:
            The computed average AP metric
        """
        ap_list: List[float] = [
            MetricsComputation.get_overall_ap(
                metrics_dict=model_metrics.metrics_dict,
                iou_threshold=iou_threshold,
            )
            for iou_threshold in model_metrics.metrics_dict.keys()
        ]
        return sum(ap_list) / len(ap_list)

    @staticmethod
    def get_ap_50_95(model_metrics: GeometricEvaluationMetrics) -> float:
        """
        Compute the COCO mAP metric which is defined as the AP
        metric for the iou-thresholds = [0.5, 0.55, ..., 0.95].

        Args:
            model_metrics: The model-metrics for which to compute the COCO mAP

        Returns:
            The computed COCO mAP
        """
        return sum(
            [
                MetricsComputation.get_overall_ap(
                    metrics_dict=model_metrics.metrics_dict, iou_threshold=iou
                )
                for iou in MetricsComputation.iou_thresholds_ap_50_95
            ]
        ) / len(MetricsComputation.iou_thresholds_ap_50_95)

    @staticmethod
    def get_ap_50(model_metrics: GeometricEvaluationMetrics) -> float:
        """
        Compute the AP50 metric which is defined as the AP
        metric for the iou-threshold=0.5

        Args:
            model_metrics: The model-metrics for which to compute the COCO mAP

        Returns:
            The computed AP50
        """

        return MetricsComputation.get_overall_ap(
            metrics_dict=model_metrics.metrics_dict, iou_threshold=0.5
        )

    def __reset_main_dictionaries(
        self,
    ) -> None:
        for class_identifier in self.class_identifier_list:
            self.model_metrics.metrics_image_info_dict[str(class_identifier)] = dict()

        for bbox_size_type in GeometricSizeTypes.get_values_as_list(class_type=GeometricSizeTypes):
            self.computing_data.gt_counter_dict[bbox_size_type] = dict()

            for class_identifier in self.class_identifier_list:
                self.computing_data.gt_counter_dict[bbox_size_type][str(class_identifier)] = 0

        for iou_thresh in self.iou_thresholds:
            self.model_metrics.metrics_dict[iou_thresh] = dict()

            for bbox_size_type in GeometricSizeTypes.get_values_as_list(
                class_type=GeometricSizeTypes
            ):
                self.model_metrics.metrics_dict[iou_thresh][bbox_size_type] = dict()

                for class_identifier in self.class_identifier_list:
                    self.model_metrics.metrics_dict[iou_thresh][bbox_size_type][
                        str(class_identifier)
                    ] = GeometricMetrics()

    def __reset_computing_dictionaries(self) -> None:
        for iou_thresh in self.iou_thresholds:
            self.computing_data.false_positives_dict[iou_thresh] = dict()
            self.computing_data.true_positives_dict[iou_thresh] = dict()
            self.computing_data.scores[iou_thresh] = dict()

            for bbox_size in GeometricSizeTypes.get_values_as_list(class_type=GeometricSizeTypes):
                self.computing_data.false_positives_dict[iou_thresh][bbox_size] = np.zeros((0,))
                self.computing_data.true_positives_dict[iou_thresh][bbox_size] = np.zeros((0,))

                self.computing_data.scores[iou_thresh][bbox_size] = np.zeros((0,))

        for bbox_size in GeometricSizeTypes.get_values_as_list(class_type=GeometricSizeTypes):
            self.computing_data.iou_dict[bbox_size] = np.zeros((0,))

    def compute_metrics(
        self,
    ) -> GeometricEvaluationMetrics:
        """
        Compute MAP Metrics

        Returns:
            The metrics dictionary in the form of:

            1st key: The iou-threshold under which the metrics should be computed
            2nd key: The type of the size of the geometric object
            3rd key: The class-identifier as string
            Value: The computed metrics

            Dict[IOU_THRESHOLD][GeometricSizeTypes.BBOX_SIZE_TYPE][CLASS_IDENTIFIER_STRING] = GeometricMetrics
        """

        self.__reset_main_dictionaries()

        for class_identifier in self.class_identifier_list:
            self.__reset_computing_dictionaries()

            self.__fill_computing_data(
                class_identifier=class_identifier,
            )

            self.__compute_metrics(class_identifier=class_identifier)

        return GeometricEvaluationMetrics(
            model_specifier=self.model_specifier,
            metrics_dict=self.get_metrics_dict(),
            metrics_image_info_dict=self.get_metrics_image_info_dict(),
        )

    def __update_from_prediction(
        self,
        index: int,
        image_path: str,
        gt_objects: List[GeometricClassification],
        predictions: List[GeometricClassification],
    ) -> None:
        """
        Update the main computing dictionaries given the information from the ground truth
        annotation and predicted bounding boxes at the dataset index.

        Args:
            index: The index of the dataset where to update the data
            image_path: The image-path for which the internal annotation data dictionaries should be updated
            predictions: The predicted perceptions for this index (data item)

        Returns:
            None
        """

        logger.debug("Get metrics for annotation: \n" "  - image-path:       %s\n" % image_path)

        self._all_gt_dict = MetricsComputation.__update_annotation_data_dict(
            class_identifier_list=self.class_identifier_list,
            all_annotations_dict=self._all_gt_dict,
            index=index,
            image_path=image_path,
            geometric_objects=gt_objects,
        )

        self._all_predictions_dict = MetricsComputation.__update_annotation_data_dict(
            class_identifier_list=self.class_identifier_list,
            all_annotations_dict=self._all_predictions_dict,
            index=index,
            image_path=image_path,
            geometric_objects=predictions,
        )

    @staticmethod
    def __update_annotation_data_dict(
        class_identifier_list: List[ClassIdentifier],
        all_annotations_dict: Dict[int, Dict[str, List[EvaluationEntry]]],
        index: int,
        image_path: str,
        geometric_objects: List[GeometricClassification],
    ) -> Dict[int, Dict[str, List[EvaluationEntry]]]:
        """
        Update the annotation data dict at the specified index with the per class geometric object
        information.

        Args:
            class_identifier_list: List of class identifier for which to update the annotation
                                   data
            all_annotations_dict: The dictionary to update
            index: int, index of entry (image) which annotations should be updated
            image_path: The image path for which the EvaluationEntries should be created

        Returns:
            The updated annotation data dict in the form of:

            1st key: The index of the dataset
            2nd key: The class-identifier as string
            Value: The list of geometrics objects that are fitting to the respective keys

            Dict[INDEX][CLASS_IDENTIFIER_STRING] = List[EvaluationEntry]
        """

        if len(geometric_objects) == 0:
            # No update necessary
            return all_annotations_dict

        # Gather (BoundingBox / Annotation) data for each class-id.
        # This is needed to be able to compute detailed metrics
        # for each class-id.
        for class_identifier in class_identifier_list:
            # Have to find all bounding-boxes for a specific class-id.
            # In order to have access to the information about the image-path,
            # we need to store an annotation object, rather than a single bounding-box.
            # Despite the fact that for the metric calculation itself the bounding-box
            # information is enough.
            # For purposes like: draw/log all false-positive bounding-boxes for specific
            # images we need to have the information about the image-path
            #   => therefore use BaseAnnotation and not BoundingBox alone
            all_annotations_dict[index][str(class_identifier)] = []
            for s in geometric_objects:
                if s.class_id == class_identifier.class_id:
                    all_annotations_dict[index][str(class_identifier)].append(
                        EvaluationEntry(
                            image_path=image_path,
                            evaluation_objects=[type(s).from_dict(s.to_dict())],  # type: ignore[list-item]
                        )
                    )

        return all_annotations_dict

    def __update_tp_fp_data(
        self, is_true_positive: bool, iou_thresh: float, bbox_size_type: str, iou_value: float
    ) -> None:
        if is_true_positive:
            tp_value = 1
            fp_value = 0
        else:
            tp_value = 0
            fp_value = 1

        # Update dictionary for overall mAP computation
        self.computing_data.true_positives_dict[iou_thresh][GeometricSizeTypes.BBOX_ALL] = (
            np.append(
                self.computing_data.true_positives_dict[iou_thresh][GeometricSizeTypes.BBOX_ALL],
                tp_value,
            )
        )
        self.computing_data.false_positives_dict[iou_thresh][GeometricSizeTypes.BBOX_ALL] = (
            np.append(
                self.computing_data.false_positives_dict[iou_thresh][GeometricSizeTypes.BBOX_ALL],
                fp_value,
            )
        )
        self.computing_data.iou_dict[GeometricSizeTypes.BBOX_ALL] = np.append(
            self.computing_data.iou_dict[GeometricSizeTypes.BBOX_ALL],
            iou_value,
        )

        # Update dictionary for size specific mAP computation
        self.computing_data.true_positives_dict[iou_thresh][bbox_size_type] = np.append(
            self.computing_data.true_positives_dict[iou_thresh][bbox_size_type],
            tp_value,
        )
        self.computing_data.false_positives_dict[iou_thresh][bbox_size_type] = np.append(
            self.computing_data.false_positives_dict[iou_thresh][bbox_size_type],
            fp_value,
        )
        self.computing_data.iou_dict[bbox_size_type] = np.append(
            self.computing_data.iou_dict[bbox_size_type],
            iou_value,
        )

    def __update_computation_dictionaries(
        self,
        prediction: GeometricClassification,
        iou_thresh: float,
        unmatched_gt_objects: List[GeometricClassification],
    ) -> List[GeometricClassification]:
        bbox_size_type: str = get_bbox_size_type(prediction.ortho_box())

        # if the ground truth data for this image indicates that nothing has to be detected,
        # indicate this box as False-Positive right away!
        if len(unmatched_gt_objects) == 0:
            self.__update_tp_fp_data(
                is_true_positive=False,
                iou_thresh=iou_thresh,
                bbox_size_type=bbox_size_type,
                iou_value=0.0,
            )
        else:
            max_overlap, assigned_evaluation_entry = compute_max_prediction(
                prediction=prediction, gt_predictions=unmatched_gt_objects
            )

            # If max-overlap fulfills given threshold
            # the detected box is treated as TP, otherwise as FP
            if (
                max_overlap
                >= iou_thresh
                # and
                # if this prediction has not already been assigned as valid,
                # treat it as true positive, otherwise as false positive
                # assigned_prediction not in self.detected_annotations[iou_thresh]
            ):
                unmatched_gt_objects.remove(assigned_evaluation_entry)

                # NOTE: In order to have correct metrics, the size of a true positive
                #       bounding box is the same as for the matching ground truth bounding box
                bbox_size_type = get_bbox_size_type(assigned_evaluation_entry.ortho_box())

                # Update as True-Positive
                self.__update_tp_fp_data(
                    is_true_positive=True,
                    iou_thresh=iou_thresh,
                    bbox_size_type=bbox_size_type,
                    iou_value=max_overlap,
                )
            else:
                # Update as False-Positive
                self.__update_tp_fp_data(
                    is_true_positive=False,
                    iou_thresh=iou_thresh,
                    bbox_size_type=bbox_size_type,
                    iou_value=0.0,
                )

        # append the score of the current bounding-box to the overall  scores list
        self.computing_data.scores[iou_thresh][GeometricSizeTypes.BBOX_ALL] = np.append(
            self.computing_data.scores[iou_thresh][GeometricSizeTypes.BBOX_ALL],
            prediction.score,
        )

        # append the score of the current bounding-box to the size specific scores list
        self.computing_data.scores[iou_thresh][bbox_size_type] = np.append(
            self.computing_data.scores[iou_thresh][bbox_size_type], prediction.score
        )

        return unmatched_gt_objects

    def __update_true_positive_metric_info(
        self,
        gt_evaluation_entry: Optional[EvaluationEntry],
        prediction_evaluation_entry: EvaluationEntry,
        prediction: GeometricClassification,
        class_identifier: ClassIdentifier,
    ) -> None:
        # The predicted annotation is an TP, therefore store it in a dict so that
        # it can be logged to tensorboard after the evaluation has finished

        # Build an EvaluationEntry based on the ground truth data and the given prediction
        tp_evaluation_entry = EvaluationEntry(
            image_path=prediction_evaluation_entry.image_path,
            evaluation_objects=[prediction],
        )

        if (
            prediction_evaluation_entry.image_path
            not in self.model_metrics.metrics_image_info_dict[str(class_identifier)]
        ):
            # There is no entry for the class-name + image-path combination yet,
            # create a new MetricImageInfo object

            self.model_metrics.metrics_image_info_dict[str(class_identifier)][
                prediction_evaluation_entry.image_path
            ] = MetricImageInfo(
                gt_evaluation_entry=gt_evaluation_entry,
                tp_evaluation_entry=tp_evaluation_entry,
            )
        else:
            # Update MetricImageInfo entry that belongs
            # to the given class-name + image-path combination
            if (
                self.model_metrics.metrics_image_info_dict[str(class_identifier)][
                    prediction_evaluation_entry.image_path
                ].tp_evaluation_entry
                is not None
            ):
                # There is already a tp_evaluation_entry, update its predictions
                #
                # NOTE: the mypy error 'Item "None" of "Optional[EvaluationEntry]" has no
                #       attribute "evaluation_objects' can be ignored. It is checked by the
                #       above query operation. Somehow mypy does not get that.
                self.model_metrics.metrics_image_info_dict[str(class_identifier)][
                    prediction_evaluation_entry.image_path  # type: ignore[union-attr]
                ].tp_evaluation_entry.evaluation_objects.append(prediction)
            else:
                # There is now tp_evaluation_entry yet, set a new one for the MetricImageInfo object
                self.model_metrics.metrics_image_info_dict[str(class_identifier)][
                    prediction_evaluation_entry.image_path
                ].tp_evaluation_entry = tp_evaluation_entry

    def __update_false_positive_metric_info(
        self,
        gt_evaluation_entry: Optional[EvaluationEntry],
        prediction_evaluation_entry: EvaluationEntry,
        prediction: GeometricClassification,
        class_identifier: ClassIdentifier,
    ) -> None:
        # The predicted annotation is an FP, therefore store it in a dict so that
        # it can be logged to tensorboard after the evaluation has finished

        # Build an EvaluationEntry based on the ground truth data and the given prediction
        fp_evaluation_entry = EvaluationEntry(
            image_path=prediction_evaluation_entry.image_path,
            evaluation_objects=[prediction],
        )

        if (
            prediction_evaluation_entry.image_path
            not in self.model_metrics.metrics_image_info_dict[str(class_identifier)]
        ):
            # There is no entry for the class-name + image-path combination yet,
            # create a new MetricImageInfo object

            self.model_metrics.metrics_image_info_dict[str(class_identifier)][
                prediction_evaluation_entry.image_path
            ] = MetricImageInfo(
                gt_evaluation_entry=gt_evaluation_entry,
                fp_evaluation_entry=fp_evaluation_entry,
            )
        else:
            # Update MetricImageInfo entry that belongs
            # to the given class-name + image-path combination
            if (
                self.model_metrics.metrics_image_info_dict[str(class_identifier)][
                    prediction_evaluation_entry.image_path
                ].fp_evaluation_entry
                is not None
            ):
                # There is already a fp_evaluation_entry, update its predictions
                #
                # NOTE: the mypy error 'Item "None" of "Optional[BaseAnnotation]" has no
                #       attribute "evaluation_objects' can be ignored. It is checked by the
                #       above query operation. Somehow mypy does not get that.
                self.model_metrics.metrics_image_info_dict[str(class_identifier)][
                    prediction_evaluation_entry.image_path  # type: ignore[union-attr]
                ].fp_evaluation_entry.evaluation_objects.append(prediction)
            else:
                # There is now fp_evaluation_entry yet, set a new one for the MetricImageInfo object
                self.model_metrics.metrics_image_info_dict[str(class_identifier)][
                    prediction_evaluation_entry.image_path
                ].fp_evaluation_entry = fp_evaluation_entry

    def __update_false_negative_metric_info(
        self,
        gt_evaluation_entry: EvaluationEntry,
        iou_unmatched_gt_objects: List[GeometricClassification],
        class_identifier: ClassIdentifier,
    ) -> None:
        """
        Update the fn_evaluation_entry entry of model_metrics.metrics_image_info_dict
        for the given class_name and the image path of the given ground truth annotation.

        Args:
            gt_evaluation_entry:
            iou_unmatched_gt_objects:
            class_identifier:

        Returns:
            None
        """

        fn_evaluation_entry = EvaluationEntry(
            image_path=gt_evaluation_entry.image_path,
            evaluation_objects=iou_unmatched_gt_objects,
        )

        if (
            gt_evaluation_entry.image_path
            not in self.model_metrics.metrics_image_info_dict[str(class_identifier)]
        ):
            self.model_metrics.metrics_image_info_dict[str(class_identifier)][
                gt_evaluation_entry.image_path
            ] = MetricImageInfo(
                gt_evaluation_entry=gt_evaluation_entry,
                fn_evaluation_entry=fn_evaluation_entry,
            )
        else:
            self.model_metrics.metrics_image_info_dict[str(class_identifier)][
                gt_evaluation_entry.image_path
            ].fn_evaluation_entry = fn_evaluation_entry

    def __get_gt_annotation(
        self, dataset_index: int, class_identifier: ClassIdentifier
    ) -> Tuple[Optional[EvaluationEntry], List[GeometricClassification]]:
        # annotation object for gathering all ground truth bounding boxes for this image
        new_evaluation_entry: Optional[EvaluationEntry] = None

        unmatched_gt_objects: List[GeometricClassification] = []

        if not str(class_identifier) in self._all_gt_dict[dataset_index]:
            return new_evaluation_entry, unmatched_gt_objects

        # Iterate over all ground-truth annotations that containing bounding-box
        # information for this image (dataset_index) and the given class-id
        for evaluation_entry in self._all_gt_dict[dataset_index][str(class_identifier)]:
            unmatched_gt_objects.extend(evaluation_entry.evaluation_objects)

            for prediction in evaluation_entry.evaluation_objects:
                # Increase ground-truth counter for overall count
                # and the specific prediction size count

                self.computing_data.gt_counter_dict[GeometricSizeTypes.BBOX_ALL][
                    str(class_identifier)
                ] += 1

                self.computing_data.gt_counter_dict[get_bbox_size_type(prediction.ortho_box())][
                    str(class_identifier)
                ] += 1

            # Initialize/update an overall ground_truth annotation that contains all
            # data for this image and class-id
            if new_evaluation_entry is None:
                new_evaluation_entry = copy.deepcopy(evaluation_entry)
            else:
                new_evaluation_entry.evaluation_objects.extend(evaluation_entry.evaluation_objects)

        return new_evaluation_entry, unmatched_gt_objects

    def __fill_computing_data(
        self,
        class_identifier: ClassIdentifier,
    ) -> None:
        process_bar = tqdm(
            range(self.dataset_length),
            desc=f"Compute metrics for class-identifier: {class_identifier}",
        )

        # Iterate over all image indices
        for dataset_index in process_bar:
            (
                gt_evaluation_entry,
                unmatched_gt_objects,
            ) = self.__get_gt_annotation(
                dataset_index=dataset_index, class_identifier=class_identifier
            )

            for iou_thresh in self.iou_thresholds:
                # Iterate over all predicted annotations for this image and class-id
                iou_unmatched_gt_objects = copy.deepcopy(unmatched_gt_objects)

                if str(class_identifier) in self._all_predictions_dict[dataset_index]:
                    for predicted_evaluation_entry in self._all_predictions_dict[dataset_index][
                        str(class_identifier)
                    ]:
                        for prediction in predicted_evaluation_entry.evaluation_objects:
                            iou_unmatched_gt_objects = self.__update_computation_dictionaries(
                                prediction=prediction,
                                iou_thresh=iou_thresh,
                                unmatched_gt_objects=iou_unmatched_gt_objects,
                            )

                            # The prediction is an FP if the last entry of the list from the
                            # false positives is equal to 1
                            if (
                                self.computing_data.false_positives_dict[iou_thresh][
                                    GeometricSizeTypes.BBOX_ALL
                                ][-1]
                                == 1
                            ):

                                self.__update_false_positive_metric_info(
                                    gt_evaluation_entry=gt_evaluation_entry,
                                    prediction_evaluation_entry=predicted_evaluation_entry,
                                    prediction=prediction,
                                    class_identifier=class_identifier,
                                )
                            else:
                                self.__update_true_positive_metric_info(
                                    gt_evaluation_entry=gt_evaluation_entry,
                                    prediction_evaluation_entry=predicted_evaluation_entry,
                                    prediction=prediction,
                                    class_identifier=class_identifier,
                                )

                # Check if there are ground truth annotations which haven't been matched,
                # this states that the box is a false negative
                if (
                    str(class_identifier) in self._all_gt_dict[dataset_index]
                    and len(self._all_gt_dict[dataset_index][str(class_identifier)]) > 0
                    and len(iou_unmatched_gt_objects) > 0
                    and gt_evaluation_entry is not None
                ):
                    self.__update_false_negative_metric_info(
                        gt_evaluation_entry=gt_evaluation_entry,
                        iou_unmatched_gt_objects=iou_unmatched_gt_objects,
                        class_identifier=class_identifier,
                    )

    def __compute_and_update_metrics_step(
        self,
        class_identifier: ClassIdentifier,
        iou_thresh: float,
        bbox_size_type: str,
        num_annotations: int,
        true_positives: np.ndarray,  # type: ignore[type-arg]
        false_positives: np.ndarray,  # type: ignore[type-arg]
    ) -> None:
        # compute recall and precision
        recall_values = true_positives / num_annotations
        precision_values = true_positives / np.maximum(
            true_positives + false_positives, np.finfo(np.float64).eps
        )

        # compute average precision
        # NOTE: mypy error 'Call to untyped function "close" in typed context' can be ignored
        ap = voc_ap(rec=recall_values, prec=precision_values, use_07_metric=False)  # type: ignore

        # IoUs only count if it is a true positive and therefore
        # has to be bigger or equal than the IoU threshold
        tp_iou_values = [
            iou_value
            for iou_value in self.computing_data.iou_dict[bbox_size_type]
            if iou_value > iou_thresh
        ]

        if len(tp_iou_values) > 0:
            avg_tp_iou = sum(tp_iou_values) / len(tp_iou_values)
        else:
            avg_tp_iou = 0.0

        if len(true_positives) > 0:
            tp = int(true_positives[-1])

            if tp > num_annotations:
                print("TP not valid")

            rc = tp / num_annotations
        else:
            tp = DEFAULT_INT_VALUE
            rc = DEFAULT_FLOAT_VALUE

        if len(false_positives) > 0:
            fp = int(false_positives[-1])
        else:
            fp = DEFAULT_INT_VALUE

        try:
            pr = tp / (fp + tp)
        except ZeroDivisionError:
            pr = DEFAULT_FLOAT_VALUE

        try:
            f1 = 2 * (pr * rc) / (pr + rc)
        except ZeroDivisionError:
            f1 = DEFAULT_FLOAT_VALUE

        fn = num_annotations - tp

        self.model_metrics.metrics_dict[iou_thresh][bbox_size_type][str(class_identifier)] = (
            GeometricMetrics(
                TP=tp,
                FP=fp,
                FN=fn,
                PR=pr,
                RC=rc,
                F1=f1,
                AP=ap,
                COUNT=num_annotations,
                AVG_TP_IOU=avg_tp_iou,
            )
        )

    def __compute_metrics(self, class_identifier: ClassIdentifier) -> None:
        for iou_thresh in self.iou_thresholds:
            for bbox_size_type in GeometricSizeTypes.get_values_as_list(
                class_type=GeometricSizeTypes
            ):
                num_annotations = self.computing_data.gt_counter_dict[bbox_size_type][
                    str(class_identifier)
                ]

                if num_annotations == 0.0:
                    continue

                # sort by score
                indices = np.argsort(-self.computing_data.scores[iou_thresh][bbox_size_type])

                false_positives = self.computing_data.false_positives_dict[iou_thresh][
                    bbox_size_type
                ][indices]

                true_positives = self.computing_data.true_positives_dict[iou_thresh][
                    bbox_size_type
                ][indices]

                # compute false positives and true positives
                false_positives = np.cumsum(false_positives)
                true_positives = np.cumsum(true_positives)

                self.__compute_and_update_metrics_step(
                    class_identifier=class_identifier,
                    iou_thresh=iou_thresh,
                    bbox_size_type=bbox_size_type,
                    num_annotations=num_annotations,
                    true_positives=true_positives,
                    false_positives=false_positives,
                )

    @staticmethod
    def match_false_negatives_and_false_positives_as_dict(
        metrics_image_info_dict: METRIC_IMAGE_INFO_TYPE,
        iou_threshold: float,
    ) -> Tuple[METRIC_IMAGE_INFO_TYPE, CONFUSION_MATRIX_DICT_TYPE]:
        """
        Build a dictionary containing the confusion matrix that is based on the given
        metrics by the metrics_image_info_dict.

        Args:
            metrics_image_info_dict: The dictionary containing the information of the metrics
            iou_threshold: The iou threshold for which to compute the confusion matrix

        Returns:
            A tuple containing:
            - The reverse metrics_image_info_dict where the keys class_identifier_str and image-path have been swapped
            - The confusion matrix
        """

        confusion_matrix: CONFUSION_MATRIX_DICT_TYPE = {
            k2: {k1: 0 for k1 in metrics_image_info_dict.keys()}
            for k2 in metrics_image_info_dict.keys()
        }

        # Build a dict for reverse look up (reverse of metrics_image_info_dict),
        # where the keys are image paths. This is needed to fill the confusion matrix per image.
        # 1st key: Image Path
        # 2nd key: Class Identifier as string
        # value: The MetricImageInfo for this image path and class name
        image_metrics_info_dict: Dict[str, Dict[str, MetricImageInfo]] = {}
        for class_identifier_str in metrics_image_info_dict.keys():
            for image_path, metric_image_info in metrics_image_info_dict[
                class_identifier_str
            ].items():
                if image_path not in image_metrics_info_dict:
                    image_metrics_info_dict[image_path] = {}

                if class_identifier_str not in image_metrics_info_dict[image_path]:
                    image_metrics_info_dict[image_path][class_identifier_str] = metric_image_info

        result_metrics_image_info_dict = copy.deepcopy(metrics_image_info_dict)

        for image_path in image_metrics_info_dict.keys():
            class_identifiers: List[ClassIdentifier] = [
                ClassIdentifier.from_str(class_identifier_str=c)
                for c in image_metrics_info_dict[image_path].keys()
            ]

            for class_identifier in class_identifiers:
                other_class_identifiers: List[ClassIdentifier] = copy.deepcopy(class_identifiers)
                other_class_identifiers.remove(class_identifier)

                class_img_metrics_info = image_metrics_info_dict[image_path][str(class_identifier)]
                if class_img_metrics_info.fn_evaluation_entry:
                    class_fn_predictions = (
                        class_img_metrics_info.fn_evaluation_entry.evaluation_objects
                    )

                    for fn_prediction in class_fn_predictions:
                        for other_class_identifier in other_class_identifiers:
                            other_class_img_metrics_info = image_metrics_info_dict[image_path][
                                str(other_class_identifier)
                            ]

                            if other_class_img_metrics_info.fp_evaluation_entry:
                                other_class_fp_predictions = (
                                    other_class_img_metrics_info.fp_evaluation_entry.evaluation_objects
                                )

                                (
                                    max_overlap,
                                    assigned_fp_prediction,
                                ) = compute_max_prediction(
                                    prediction=fn_prediction,
                                    gt_predictions=other_class_fp_predictions,
                                )

                                if max_overlap > iou_threshold:
                                    confusion_matrix[str(class_identifier)][
                                        str(other_class_identifier)
                                    ] += 1
                                    class_img_metrics_info.fn_evaluation_entry.evaluation_objects.remove(
                                        fn_prediction
                                    )
                                    other_class_img_metrics_info.fp_evaluation_entry.evaluation_objects.remove(
                                        assigned_fp_prediction
                                    )

                                    result_metrics_image_info_dict = MetricsComputation.__change_matched_bbox_attribute(
                                        metrics_image_info_dict=result_metrics_image_info_dict,
                                        image_path=class_img_metrics_info.fn_evaluation_entry.image_path,
                                        class_identifier=class_identifier,
                                        other_class_identifier=other_class_identifier,
                                        fn_prediction=fn_prediction,
                                        assigned_fp_prediction=assigned_fp_prediction,
                                        class_img_metrics_info=class_img_metrics_info,
                                        other_class_img_metrics_info=other_class_img_metrics_info,
                                    )

                            if other_class_img_metrics_info.fp_evaluation_entry:
                                if (
                                    len(
                                        other_class_img_metrics_info.fp_evaluation_entry.evaluation_objects
                                    )
                                    == 0
                                ):
                                    other_class_img_metrics_info.fp_evaluation_entry = None

                if class_img_metrics_info.fn_evaluation_entry:
                    if len(class_img_metrics_info.fn_evaluation_entry.evaluation_objects) == 0:
                        class_img_metrics_info.fn_evaluation_entry = None

        return result_metrics_image_info_dict, confusion_matrix

    @staticmethod
    def __change_matched_bbox_attribute(
        metrics_image_info_dict: METRIC_IMAGE_INFO_TYPE,
        image_path: str,
        class_identifier: ClassIdentifier,
        other_class_identifier: ClassIdentifier,
        fn_prediction: GeometricClassification,
        assigned_fp_prediction: GeometricClassification,
        class_img_metrics_info: MetricImageInfo,
        other_class_img_metrics_info: MetricImageInfo,
    ) -> METRIC_IMAGE_INFO_TYPE:
        if class_img_metrics_info.fn_matched_fp_evaluation_entry is not None:
            class_img_metrics_info.fn_matched_fp_evaluation_entry.evaluation_objects.append(
                fn_prediction
            )
        else:
            class_img_metrics_info.fn_matched_fp_evaluation_entry = EvaluationEntry(
                image_path=image_path,
                evaluation_objects=[fn_prediction],
            )

        if other_class_img_metrics_info.fn_matched_fp_evaluation_entry is not None:
            other_class_img_metrics_info.fn_matched_fp_evaluation_entry.evaluation_objects.append(
                assigned_fp_prediction
            )

        else:
            other_class_img_metrics_info.fn_matched_fp_evaluation_entry = EvaluationEntry(
                image_path=image_path,
                evaluation_objects=[assigned_fp_prediction],
            )

        metrics_image_info_dict[str(class_identifier)][image_path] = class_img_metrics_info
        metrics_image_info_dict[str(other_class_identifier)][
            image_path
        ] = other_class_img_metrics_info

        return metrics_image_info_dict
